import itertools
from contextlib import nullcontext as does_not_raise
from unittest.mock import patch

import numpy as np
import pynapple as nap
import pytest
from conftest import (
    SizeTerminal,
    basis_collapse_all_non_vec_axis,
    basis_with_add_kwargs,
    custom_basis,
    custom_basis_2d,
)
from numpy.typing import NDArray

from nemos.basis._custom_basis import CustomBasis, apply_f_vectorized


@pytest.fixture
def vectorize_func(request):
    ndim = request.param

    def not_vectorized_func(*xi) -> NDArray:
        if any(x.ndim != ndim for x in xi):
            raise ValueError("Non-vectorized function")
        # sum over second dimension and add results
        return sum(np.sum(x.reshape(x.shape[0], -1), axis=1) for x in xi)

    return (
        lambda *xi: apply_f_vectorized(not_vectorized_func, *xi, ndim_input=ndim),
        not_vectorized_func,
    )


@pytest.fixture
def vec_func_kwargs(request):
    kwargs = request.param

    def func(*xi, add=0, mul=1) -> NDArray:
        return xi[0] * mul + add

    return lambda *xi, **k: apply_f_vectorized(func, *xi, ndim_input=1, **k), kwargs


@pytest.mark.parametrize(
    "vectorize_func, x, expectation",
    [
        (
            1,
            np.random.randn(
                10,
            ),
            does_not_raise(),
        ),
        (
            1,
            np.random.randn(10, 2),
            pytest.raises(ValueError, match="Non-vectorized function"),
        ),
        (2, np.random.randn(10, 2), does_not_raise()),
        (
            2,
            np.random.randn(10, 2, 1),
            pytest.raises(ValueError, match="Non-vectorized function"),
        ),
    ],
    indirect=["vectorize_func"],
)
def test_vec_function_dims(vectorize_func, x, expectation):
    vec_f, regular_f = vectorize_func
    # should be always fine
    try:
        out_vec = vec_f(x)
    except Exception:
        raise ValueError("Failed vectorization.")
    with expectation:
        out_reg = regular_f(x)
        assert np.all(out_vec.flatten() == out_reg.flatten())


@pytest.mark.parametrize(
    "vectorize_func, x, expected_out_shape",
    [
        (
            1,
            (
                np.random.randn(
                    10,
                ),
            ),
            (10, 1),
        ),
        # vectorized over second dimension
        (1, (np.random.randn(10, 2),), (10, 2)),
        # no vectorized dimension (mean over all axis except first)
        (2, (np.random.randn(10, 2),), (10, 1)),
        # no vectorized dimension (mean over all axis except first)
        (2, (np.random.randn(10, 2), np.random.randn(10, 3)), (10, 1)),
        # vectorize 3rd dimension
        (2, (np.random.randn(10, 2, 3), np.random.randn(10, 2, 3)), (10, 3)),
    ],
    indirect=["vectorize_func"],
)
def test_vec_function_output_shape(vectorize_func, x, expected_out_shape):
    vec_f, regular_f = vectorize_func
    out = vec_f(*x)
    assert out.shape == expected_out_shape


@pytest.mark.parametrize(
    "x_shape",
    [
        (10, 3),  # 2D inputs (1D vectorization)
        (10, 3, 4),  # 3D inputs (2D vectorization)
        (10, 2, 3, 4),  # 4D inputs (3D vectorization)
    ],
)
@pytest.mark.parametrize(
    "vectorize_func, ndim",
    [(1, 1), (2, 2)],
    indirect=["vectorize_func"],
)
def test_vectorization_equivalence(x_shape, vectorize_func, ndim):
    """Test that vectorized computation equals explicit nested loops."""
    if len(x_shape) <= ndim:
        pytest.skip("input shape should be > ndim")
    # Seed for reproducibility
    np.random.seed(42)

    # Create random inputs
    xis = [np.random.randn(*x_shape) for x in range(ndim)]
    n_basis_funcs = 1

    # Get vectorized result
    vec_f, regular_f = vectorize_func
    vectorized_result = vec_f(*xis)

    # Compute expected result with explicit loops
    n_samples = x_shape[0]
    vec_shape = x_shape[ndim:]  # vectorized dimensions

    # Initialize output array
    out = np.empty((xis[0].shape[0], *vec_shape, n_basis_funcs))

    # Generate all combinations of vectorized indices
    vec_indices = itertools.product(*[range(dim) for dim in vec_shape])

    for indices in vec_indices:
        # Extract 1D slices for this combination of indices
        x_slices = [
            xi[(slice(None),) * ndim + indices] for xi in xis
        ]  # x[:, i, j, ...]

        # Compute features for this slice
        slice_result = regular_f(*x_slices)

        # Store in output array
        out[(slice(None),) + indices + (slice(None),)] = (
            slice_result[:, None] if slice_result.ndim == 1 else slice_result
        )

        # Reshape to match expected output format: (n_samples, flattened_features)
        expected_result = out.reshape(n_samples, -1)

    # Verify equivalence
    np.testing.assert_array_equal(vectorized_result, expected_result)

    # Also verify shapes are correct
    expected_n_features = (
        np.prod(vec_shape) * n_basis_funcs if vec_shape else n_basis_funcs
    )
    assert vectorized_result.shape == (n_samples, expected_n_features)


@pytest.mark.parametrize(
    "vec_func_kwargs, x",
    [
        (dict(add=1), np.zeros((10, 2))),
        (dict(add=0), np.zeros((10, 2))),
        (dict(add=1, mul=2), np.ones((10, 2))),
        (dict(mul=0), np.ones((10, 2))),
        ({}, np.ones((10, 2))),
    ],
    indirect=["vec_func_kwargs"],
)
def test_vec_function_kwargs(vec_func_kwargs, x):
    vec_f, kwargs = vec_func_kwargs
    out = vec_f(x, **kwargs)
    add = kwargs.get("add", 0)
    mul = kwargs.get("mul", 1)
    assert np.all(out == x * mul + add)


@pytest.mark.parametrize("n_basis_funcs", [4])
@pytest.mark.parametrize(
    "input_dim, expectation",
    [
        (1, does_not_raise()),
        (2, pytest.raises(ValueError, match="Each input must have at least")),
    ],
)
def test_input_dimension(n_basis_funcs, input_dim, expectation):
    bas = custom_basis(n_basis_funcs=n_basis_funcs, label=None, ndim_input=input_dim)
    with expectation:
        shape = (10, *(1 for _ in range(1 - 1)))
        bas.compute_features(*shape)


@pytest.mark.parametrize("n_basis_funcs", [4, 5])
def test_custom_basis_feature_1d_shape(n_basis_funcs):
    bas = custom_basis(n_basis_funcs=n_basis_funcs, label=None)
    out = bas.compute_features(
        np.random.randn(
            10,
        )
    )
    assert out.shape == (10, n_basis_funcs)
    out = bas.compute_features(np.random.randn(10, 2, 3))
    assert out.shape == (10, 6 * n_basis_funcs)


@pytest.mark.parametrize("n_basis_funcs", [4, 5])
def test_custom_basis_feature_2d_shape(n_basis_funcs):
    bas = custom_basis(n_basis_funcs=n_basis_funcs, label=None, ndim_input=1)
    out = bas.compute_features(
        np.random.randn(
            10,
        )
    )
    assert out.shape == (10, n_basis_funcs)
    out = bas.compute_features(np.random.randn(10, 2, 3, 2))
    assert out.shape == (10, 12 * n_basis_funcs)

    # define a basis that behaves differently on vec
    bas = basis_collapse_all_non_vec_axis(n_basis_funcs, ndim_input=1)
    # vec second dim
    out = bas.compute_features(np.random.randn(10, 2))
    assert out.shape == (10, n_basis_funcs * 2)
    out = bas.compute_features(np.random.randn(10, 2, 2))
    assert out.shape == (10, n_basis_funcs * 4)


@pytest.mark.parametrize(
    "out_shape, expectation",
    [
        (1, does_not_raise()),
        (-1, pytest.raises(ValueError, match="Output shape must be strictly")),
        (2, does_not_raise()),
        ((1, 2), does_not_raise()),
        (0.5, pytest.raises(TypeError, match="`output_shape` must be an iterable of")),
        (
            ("a", 2),
            pytest.raises(
                ValueError, match="The tuple provided contains non integer values"
            ),
        ),
    ],
)
def test_output_shape_setter(out_shape, expectation):
    with expectation:
        custom_basis(n_basis_funcs=10, output_shape=out_shape, ndim_input=1)


def test_output_shape_reset():
    bas = custom_basis(n_basis_funcs=3, output_shape=(2, 3, 4), ndim_input=1)
    assert bas.output_shape == (2, 3, 4)
    bas.compute_features(np.linspace(0, 1, 10))
    # the output shape is set to empty tuple because each basis function
    # returns a flat array
    assert bas.output_shape == ()


@pytest.mark.parametrize("kwargs", [dict(add=0), dict(add=1), dict(), None])
def test_kwargs_apply(kwargs):
    bas = basis_with_add_kwargs(basis_kwargs=kwargs)
    if kwargs:
        assert np.all(
            np.full((10, 1), kwargs["add"]) == bas.compute_features(np.zeros(10))
        )
    else:
        assert bas.basis_kwargs == {}


@pytest.mark.parametrize(
    "kwargs, expectation",
    [
        (dict(add=0), does_not_raise()),
        (1, pytest.raises(ValueError, match="`basis_kwargs` must be a dictionary")),
        (dict(), does_not_raise()),
        (None, does_not_raise()),
        (dict(ndim_input=1), pytest.raises(ValueError, match="Invalid kwargs name")),
    ],
)
def test_basis_kwargs_set(kwargs, expectation):
    with expectation:
        bas = basis_with_add_kwargs(basis_kwargs=kwargs)
        bas.basis_kwargs == kwargs
    bas = basis_with_add_kwargs(basis_kwargs=None)
    with expectation:
        bas.basis_kwargs = kwargs
        bas.basis_kwargs == kwargs


def test_pynapple_support():
    bas = custom_basis(5, pynapple_support=True)
    x = nap.Tsd(np.arange(10), np.linspace(0, 1, 10))
    assert isinstance(bas.compute_features(x), nap.TsdFrame)
    bas = custom_basis(5, pynapple_support=False)
    with pytest.raises(TypeError):
        bas.compute_features(x)


@pytest.mark.parametrize("ps", [True, False, None, 1, 0, -1, np.array([1, 2])])
def test_pynapple_support_type(ps):
    """Test that any value that can be parsed to bool is valid."""
    if isinstance(ps, np.ndarray):
        expect = pytest.raises(ValueError, match="The truth")
    else:
        expect = does_not_raise()
    with expect:
        b = custom_basis(5, pynapple_support=ps)
        assert isinstance(b.pynapple_support, bool)


@pytest.mark.parametrize(
    "inp_dim, vec_shape, expected_num",
    [(1, [1], 5), (1, [], 5), (2, [], 10), (2, [3], 30), (2, [3, 2], 60)],
)
def test_n_output_features_match(inp_dim, expected_num, vec_shape):
    shape = [10] + [2] * (inp_dim - 1) + vec_shape
    x = np.random.randn(*shape)
    bas = custom_basis(5, ndim_input=inp_dim)
    out = bas.compute_features(x)
    assert out.shape[1] == bas.n_output_features == expected_num


@pytest.mark.parametrize(
    "inp_dim, vec_shape, expected_out_shape",
    [
        (1, [], (10, 5)),
        (1, [1], (10, 1, 5)),
        (2, [], (10, 2, 5)),
        (2, [1], (10, 2, 1, 5)),
        (2, [3], (10, 2, 3, 5)),
        (2, [3, 2], (10, 2, 6, 5)),
    ],
)
def test_features_match_evaluate(inp_dim, vec_shape, expected_out_shape):
    shape = [10] + [2] * (inp_dim - 1) + vec_shape
    x = np.random.randn(*shape)
    bas = custom_basis(5, ndim_input=inp_dim)
    out = bas.evaluate(x)
    assert out.shape == expected_out_shape


@pytest.mark.parametrize("ps", [True, False])
def test_basis_repr(ps):
    """Check that repr strips the expectation"""
    with patch("os.get_terminal_size", return_value=SizeTerminal(80, 24)):
        bas = custom_basis(5, pynapple_support=ps)
        assert (
            repr(bas)
            == f"CustomBasis(\n    funcs=[partial(power_func, 1), ..., partial(power_func, 5)],\n    ndim_input=1,\n    pynapple_support={ps},\n    is_complex=False\n)"
        )
        bas = custom_basis(1, pynapple_support=ps)
        assert (
            repr(bas)
            == f"CustomBasis(\n    funcs=[partial(power_func, 1)],\n    ndim_input=1,\n    pynapple_support={ps},\n    is_complex=False\n)"
        )
        # check composite basis repr
        assert (
            repr(bas + bas)
            == f"'(CustomBasis + CustomBasis_1)': AdditiveBasis(\n    basis1=CustomBasis(\n        funcs=[partial(power_func, 1)],\n        ndim_input=1,\n        pynapple_support={ps},\n        is_complex=False\n    ),\n    basis2='CustomBasis_1': CustomBasis(\n        funcs=[partial(power_func, 1)],\n        ndim_input=1,\n        pynapple_support={ps},\n        is_complex=False\n    ),\n)"
        )


@pytest.mark.parametrize("input_shape", [(1,), (1, 2), (1, 2, 3), ()])
def test_split_by_features_shape(input_shape):
    bas = custom_basis(4)
    out = bas.compute_features(np.random.randn(10, *input_shape))
    split = bas.split_by_feature(out, axis=1)["CustomBasis"]
    assert split.shape == (10, *input_shape, 4)


@pytest.mark.parametrize(
    "ishape, n_out_features, expectation",
    [
        ((1, 1), 5, does_not_raise()),
        (
            (1, 2),
            10,
            pytest.raises(ValueError, match="CustomBasis requires all inputs"),
        ),
        (
            (2, 1),
            10,
            pytest.raises(ValueError, match="CustomBasis requires all inputs"),
        ),
        ((2, 2), 10, does_not_raise()),
        (
            ((2, 2), 1),
            20,
            pytest.raises(ValueError, match="CustomBasis requires all inputs"),
        ),
        (
            (1, (2, 2)),
            20,
            pytest.raises(ValueError, match="CustomBasis requires all inputs"),
        ),
        (((2, 2), (2, 2)), 20, does_not_raise()),
    ],
)
def test_set_input_shape_2d(ishape, n_out_features, expectation):
    """Test that the output features match expectation when setting input shape.

    Note that the 1D case is tested in test_basis.py::TestSharedMethods.
    """
    bas = custom_basis_2d(5)
    with expectation:
        bas.set_input_shape(*ishape)
        assert bas.n_output_features == n_out_features


def test_inconsistent_input_num():
    # define a list of funcs with inconsistent input number
    invalid_funcs = [lambda x: x, lambda x, y: x]
    with pytest.raises(
        ValueError, match="Each function provided to ``funcs`` in ``CustomBasis``"
    ):
        CustomBasis(invalid_funcs)


@pytest.mark.parametrize(
    "x_shape",
    [
        (10, 3),  # 2D inputs (1D vectorization)
        (10, 3, 4),  # 3D inputs (2D vectorization)
        (10, 2, 3, 4),  # 4D inputs (3D vectorization)
    ],
)
@pytest.mark.parametrize(
    "vectorize_func, ndim",
    [(1, 1), (2, 2)],
    indirect=["vectorize_func"],
)
def test_vectorization_equivalence_basis(x_shape, vectorize_func, ndim):
    """Test that vectorized computation equals explicit nested loops."""
    if len(x_shape) <= ndim:
        pytest.skip("input shape should be > ndim")

    # define a basis
    def power_func(x, n=1):
        return np.sum(np.power(x, n), axis=1) if x.ndim > 1 else np.power(x, n)

    from functools import partial

    funcs = [partial(power_func, n=n) for n in range(1, 4 + 1)]
    bas = CustomBasis(funcs, ndim_input=ndim)
    # Seed for reproducibility
    np.random.seed(42)

    # Create random inputs
    xis = [np.random.randn(*x_shape)]
    n_basis_funcs = bas.n_basis_funcs

    # Get vectorized result
    regular_fs = bas.funcs
    vectorized_result = bas.compute_features(*xis)

    # Compute expected result with explicit loops
    n_samples = x_shape[0]
    vec_shape = x_shape[ndim:]  # vectorized dimensions

    # Initialize output array
    out = np.empty((xis[0].shape[0], *vec_shape, n_basis_funcs))

    # Generate all combinations of vectorized indices
    vec_indices = itertools.product(*[range(dim) for dim in vec_shape])

    for indices in vec_indices:
        # Extract 1D slices for this combination of indices
        x_slices = [
            xi[(slice(None),) * ndim + indices] for xi in xis
        ]  # x[:, i, j, ...]

        # Compute features for this slice
        slice_result = np.stack([f(*x_slices) for f in regular_fs], axis=-1)

        # Store in output array
        out[(slice(None),) + indices + (slice(None),)] = (
            slice_result[:, None] if slice_result.ndim == 1 else slice_result
        )

    # Reshape to match expected output format: (n_samples, flattened_features)
    expected_result = out.reshape(n_samples, -1)

    # Verify equivalence
    np.testing.assert_array_equal(vectorized_result, expected_result)

    # Also verify shapes are correct
    expected_n_features = (
        np.prod(vec_shape) * n_basis_funcs if vec_shape else n_basis_funcs
    )
    assert vectorized_result.shape == (n_samples, expected_n_features)



import itertools
from functools import partial

import jax
import numpy as np
import pytest

from nemos.fetch import fetch_data
from nemos.glm import GLM
from nemos.glm_hmm.expectation_maximization import (
    backward_pass,
    compute_xi,
    forward_backward,
    forward_pass,
    hmm_negative_log_likelihood,
    run_m_step,
)
from nemos.observation_models import BernoulliObservations, PoissonObservations
from nemos.third_party.jaxopt.jaxopt import LBFGS


def prepare_partial_hmm_nll_single_neuron(obs):
    # Define nll vmap function
    negative_log_likelihood = jax.vmap(
        lambda x, z: obs._negative_log_likelihood(
            x, z, aggregate_sample_scores=lambda w: w
        ),
        in_axes=(None, 1),
        out_axes=1,
    )

    # Solver
    def partial_hmm_negative_log_likelihood(
        weights, design_matrix, observations, posterior_prob
    ):
        return hmm_negative_log_likelihood(
            weights,
            X=design_matrix,
            y=observations,
            posteriors=posterior_prob,
            inverse_link_function=obs.default_inverse_link_function,
            negative_log_likelihood_func=negative_log_likelihood,
        )

    solver = LBFGS(partial_hmm_negative_log_likelihood, tol=10**-13)

    return partial_hmm_negative_log_likelihood, solver


def prepare_gammas_and_xis_for_m_step_single_neuron(
    X, y, initial_prob, transition_prob, glm_params, new_sess, obs
):
    (coef, intercept) = glm_params
    likelihood = jax.vmap(
        lambda x, z: obs.likelihood(x, z, aggregate_sample_scores=lambda w: w),
        in_axes=(None, 1),
        out_axes=1,
    )
    gammas, xis, _, _, _, _ = forward_backward(
        X,
        y,
        initial_prob,
        transition_prob,
        (coef, intercept),
        likelihood_func=likelihood,
        inverse_link_function=obs.default_inverse_link_function,
        is_new_session=new_sess.astype(bool),
    )

    return gammas, xis


def forward_step_numpy(py_z, new_sess, initial_prob, transition_prob):
    n_time_bins, n_states = py_z.shape
    alphas = np.full((n_time_bins, n_states), np.nan)
    c = np.full(n_time_bins, np.nan)
    for t in range(n_time_bins):
        if new_sess[t]:
            alphas[t] = initial_prob * py_z[t]
        else:
            alphas[t] = py_z[t] * (transition_prob.T @ alphas[t - 1])

        c[t] = np.sum(alphas[t])
        alphas[t] /= c[t]
    return alphas, c


def backward_step_numpy(py_z, c, new_sess, transition_prob):
    n_time_bins, n_states = py_z.shape
    betas = np.full((n_time_bins, n_states), np.nan)
    betas[-1] = np.ones(n_states)

    for t in range(n_time_bins - 2, -1, -1):
        if new_sess[t + 1]:
            betas[t] = np.ones(n_states)
        else:
            betas[t] = transition_prob @ (betas[t + 1] * py_z[t + 1])
            betas[t] /= c[t + 1]
    return betas


# Below are tests for Bernoulli observation 3 states model glm hmm
@pytest.mark.parametrize(
    "decorator",
    [
        lambda x: x,
        partial(jax.jit, static_argnames=["likelihood_func", "inverse_link_function"]),
    ],
)
def test_forward_backward_regression(decorator):
    jax.config.update("jax_enable_x64", True)

    # Fetch the data
    data_path = fetch_data("em_three_states.npz")
    data = np.load(data_path)

    X, y = data["X"], data["y"]
    new_sess = data["new_sess"]

    # assert first column is intercept, otherwise this will break
    # nemos assumption (existence of an intercept term)
    np.testing.assert_array_equal(X[:, 0], 1)

    # E-step initial parameters
    initial_prob = data["initial_prob"]
    intercept, coef = data["projection_weights"][:1], data["projection_weights"][1:]
    transition_prob = data["transition_prob"]

    # E-step output
    xis = data["xis"]
    gammas = data["gammas"]
    ll_orig, ll_norm_orig = data["log_likelihood"], data["ll_norm"]
    alphas, betas = data["alphas"], data["betas"]

    obs = BernoulliObservations()

    likelihood = jax.vmap(
        lambda x, z: obs.likelihood(x, z, aggregate_sample_scores=lambda w: w),
        in_axes=(None, 1),
        out_axes=1,
    )

    decorated_forward_backward = decorator(forward_backward)
    gammas_nemos, xis_nemos, ll_nemos, ll_norm_nemos, alphas_nemos, betas_nemos = (
        decorated_forward_backward(
            X[:, 1:],  # drop intercept
            y,
            initial_prob,
            transition_prob,
            (coef, intercept),
            likelihood_func=likelihood,
            inverse_link_function=obs.default_inverse_link_function,
            is_new_session=new_sess.astype(bool),
        )
    )

    # First testing alphas and betas because they are computed first
    np.testing.assert_almost_equal(alphas_nemos, alphas, decimal=8)
    np.testing.assert_almost_equal(betas_nemos, betas, decimal=8)

    # testing log likelihood and normalized log likelihood
    np.testing.assert_almost_equal(ll_nemos, ll_orig, decimal=8)
    np.testing.assert_almost_equal(ll_norm_nemos, ll_norm_orig, decimal=8)

    # Next testing xis and gammas because they depend on alphas and betas
    # Testing Eq. 13.43 of Bishop
    np.testing.assert_almost_equal(gammas_nemos, gammas, decimal=8)
    # Testing Eq. 13.65 of Bishop
    np.testing.assert_almost_equal(xis_nemos, xis, decimal=8)


@pytest.fixture(scope="module")
def generate_data_multi_state():
    np.random.seed(44)
    jax.config.update("jax_enable_x64", True)

    # E-step initial parameters
    n_states, n_samples = 5, 100
    initial_prob = np.random.uniform(size=(n_states))
    initial_prob /= np.sum(initial_prob)
    transition_prob = np.random.uniform(size=(n_states, n_states))
    transition_prob /= np.sum(transition_prob, axis=0)
    transition_prob = transition_prob.T
    coef, intercept = np.random.randn(2, n_states), np.random.randn(n_states)

    X = np.random.randn(n_samples, 2)
    y = np.zeros(n_samples)
    for i, k in enumerate(range(0, 100, 10)):
        sl = slice(k, k + 10)
        state = i % n_states
        rate = np.exp(X[sl].dot(coef[:, state]) + intercept[state])
        y[sl] = np.random.poisson(rate)

    new_sess = np.zeros(n_samples)
    new_sess[[0, 10, 90]] = 1
    return new_sess, initial_prob, transition_prob, coef, intercept, X, y


@pytest.fixture(scope="module")
def generate_data_multi_state_population():
    np.random.seed(44)
    jax.config.update("jax_enable_x64", True)

    # E-step initial parameters
    n_states, n_neurons, n_samples = 5, 3, 100
    initial_prob = np.random.uniform(size=(n_states))
    initial_prob /= np.sum(initial_prob)
    transition_prob = np.random.uniform(size=(n_states, n_states))
    transition_prob /= np.sum(transition_prob, axis=0)
    transition_prob = transition_prob.T
    coef, intercept = np.random.randn(2, n_neurons, n_states), np.random.randn(
        n_neurons, n_states
    )

    X = np.random.randn(n_samples, 2)
    y = np.zeros((n_samples, n_neurons))
    for i, k in enumerate(range(0, 100, 10)):
        sl = slice(k, k + 10)
        state = i % n_states
        rate = np.exp(X[sl].dot(coef[..., state]) + intercept[..., state])
        y[sl] = np.random.poisson(rate)

    new_sess = np.zeros(n_samples)
    new_sess[[0, 10, 90]] = 1
    return new_sess, initial_prob, transition_prob, coef, intercept, X, y


def test_for_loop_forward_step(generate_data_multi_state):
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state
    )

    obs = PoissonObservations()

    likelihood = jax.vmap(
        lambda x, z: obs.likelihood(x, z, aggregate_sample_scores=lambda w: w),
        in_axes=(None, 1),
        out_axes=1,
    )

    predicted_rate_given_state = obs.default_inverse_link_function(X @ coef + intercept)
    conditionals = likelihood(y, predicted_rate_given_state)

    alphas, normalization = forward_pass(
        initial_prob, transition_prob, conditionals, new_sess
    )

    alphas_numpy, normalization_numpy = forward_step_numpy(
        conditionals, new_sess, initial_prob, transition_prob
    )
    np.testing.assert_almost_equal(alphas_numpy, alphas)
    np.testing.assert_almost_equal(normalization_numpy, normalization)


def test_for_loop_backward_step(generate_data_multi_state):
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state
    )
    obs = PoissonObservations()

    likelihood = jax.vmap(
        lambda x, z: obs.likelihood(x, z, aggregate_sample_scores=lambda w: w),
        in_axes=(None, 1),
        out_axes=1,
    )

    predicted_rate_given_state = obs.default_inverse_link_function(X @ coef + intercept)
    conditionals = likelihood(y, predicted_rate_given_state)

    alphas, normalization = forward_pass(
        initial_prob, transition_prob, conditionals, new_sess
    )

    betas = backward_pass(transition_prob, conditionals, normalization, new_sess)
    betas_numpy = backward_step_numpy(
        conditionals, normalization, new_sess, transition_prob
    )
    np.testing.assert_almost_equal(betas_numpy, betas)


@pytest.fixture
def single_state_inputs():
    np.random.seed(42)
    initial_prob = np.ones(1)
    transition_prob = np.ones((1, 1))
    coef, intercept = np.random.randn(2, 1), np.random.randn(1)
    X = np.random.randn(10, 2)
    rate = np.exp(X.dot(coef) + intercept)
    y = np.random.poisson(rate[:, 0])
    return initial_prob, transition_prob, coef, intercept, X, rate, y


def test_single_state_estep(single_state_inputs):
    """Single state forward pass posteriors reduces to ones (there is a single state)."""

    initial_prob, transition_prob, coef, intercept, X, rate, y = single_state_inputs
    obs = PoissonObservations()

    likelihood = jax.vmap(
        lambda x, z: obs.likelihood(x, z, aggregate_sample_scores=lambda w: w),
        in_axes=(None, 1),
        out_axes=1,
    )
    conditionals = likelihood(y, rate)
    new_sess = np.zeros(10)
    new_sess[0] = 1
    alphas, norm = forward_pass(initial_prob, transition_prob, conditionals, new_sess)
    betas = backward_pass(transition_prob, conditionals, norm, new_sess)

    # check that the normalization factor reduces to the p(x_t | z_t)
    np.testing.assert_array_almost_equal(norm, conditionals[:, 0])

    # Note: alphas * betas is p(z_t | X), so it's automatically ones if the
    # two assertions passes, no need to check explicitly for  p(z_t | X).
    np.testing.assert_array_almost_equal(np.ones_like(alphas), alphas)
    np.testing.assert_array_almost_equal(np.ones_like(betas), betas)

    # xis are a sum of the ones over valid entires
    xis = compute_xi(
        alphas,
        betas,
        conditionals,
        norm,
        new_sess,
        transition_prob,
    )
    np.testing.assert_array_almost_equal(
        np.array([[alphas.shape[0] - sum(new_sess)]]).astype(xis), xis
    )


@pytest.mark.parametrize(
    "decorator",
    [
        lambda x: x,
        partial(
            jax.jit,
            static_argnames=["negative_log_likelihood_func", "inverse_link_function"],
        ),
    ],
)
def test_hmm_negative_log_likelihood_regression(decorator):
    jax.config.update("jax_enable_x64", True)

    # Fetch the data
    data_path = fetch_data("em_three_states.npz")
    data = np.load(data_path)

    # Design matrix and observed choices
    X, y = data["X"], data["y"]

    # Likelihood input
    gammas = data["gammas"]
    projection_weights = data["projection_weights_nll"]
    intercept, coef = projection_weights[:1], projection_weights[1:]

    # Negative LL output
    nll_m_step = data["nll_m_step"]

    # Initialize nemos observation model
    obs = BernoulliObservations()

    # Define negative log likelihood vmap function
    negative_log_likelihood = jax.vmap(
        lambda x, z: obs._negative_log_likelihood(
            x, z, aggregate_sample_scores=lambda w: w
        ),
        in_axes=(None, 1),
        out_axes=1,
    )

    nll_m_step_nemos = hmm_negative_log_likelihood(
        (coef, intercept),
        X[:, 1:],  # drop intercept column
        y,
        gammas,
        inverse_link_function=obs.default_inverse_link_function,
        negative_log_likelihood_func=negative_log_likelihood,
    )

    # Testing output of negative log likelihood
    np.testing.assert_almost_equal(nll_m_step_nemos, nll_m_step)


def expected_log_likelihood_wrt_transitions(
    transition_prob, xis, dirichlet_alphas=None
):
    likelihood = jax.numpy.sum(xis * jax.numpy.log(transition_prob))
    if dirichlet_alphas is None:
        return likelihood
    prior = (jax.numpy.log(transition_prob) * (dirichlet_alphas - 1)).sum()
    return likelihood + prior


def expected_log_likelihood_wrt_initial_prob(
    initial_prob, gammas, dirichlet_alphas=None
):
    likelihood = jax.numpy.sum(gammas * jax.numpy.log(initial_prob))
    if dirichlet_alphas is None:
        return likelihood
    prior = (jax.numpy.log(initial_prob) * (dirichlet_alphas - 1)).sum()
    return likelihood + prior


def lagrange_mult_loss(param, args, loss, **kwargs):
    proba, lam = param
    n_states = proba.shape[0]
    if proba.ndim == 2:
        constraint = proba.sum(axis=1) - jax.numpy.ones(n_states)
    else:
        constraint = proba.sum() - 1
    lagrange_mult_term = (lam * constraint).sum()
    return loss(proba, args, **kwargs) + lagrange_mult_term


def test_run_m_step_regression():
    jax.config.update("jax_enable_x64", True)

    # Fetch the data
    data_path = fetch_data("em_three_states.npz")
    data = np.load(data_path)

    # Design matrix and observed choices
    X, y = data["X"], data["y"]

    # M-step input
    gammas = data["gammas"]
    xis = data["xis"]
    projection_weights = data["projection_weights"]
    intercept, coef = projection_weights[:1], projection_weights[1:]
    new_sess = data["new_sess"]

    # M-step output
    optimized_projection_weights = data["optimized_projection_weights"]
    opt_intercept, opt_coef = (
        optimized_projection_weights[:1],
        optimized_projection_weights[1:],
    )
    new_initial_prob = data["new_initial_prob"]
    new_transition_prob = data["new_transition_prob"]

    # Initialize nemos observation model
    obs = BernoulliObservations()

    partial_hmm_negative_log_likelihood, solver = prepare_partial_hmm_nll_single_neuron(
        obs
    )

    (
        optimized_projection_weights_nemos,
        new_initial_prob_nemos,
        new_transition_prob_nemos,
        state,
    ) = run_m_step(
        X[:, 1:],  # drop intercept column
        y,
        gammas,
        xis,
        (coef, intercept),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
    )

    n_ll_nemos = partial_hmm_negative_log_likelihood(
        optimized_projection_weights_nemos,
        X[:, 1:],
        y,
        gammas,
    )
    n_ll_original = partial_hmm_negative_log_likelihood(
        (opt_coef, opt_intercept),
        X[:, 1:],
        y,
        gammas,
    )

    # Testing Eq. 13.18 of Bishop
    np.testing.assert_almost_equal(new_initial_prob_nemos, new_initial_prob)
    # Testing Eq. 13.19 of Bishop

    np.testing.assert_almost_equal(new_transition_prob_nemos, new_transition_prob)

    # Testing output of negative log likelihood
    np.testing.assert_almost_equal(n_ll_original, n_ll_nemos, decimal=10)

    jax.tree_util.tree_map(
        lambda x, y: np.testing.assert_almost_equal(x, y, decimal=6),
        (opt_coef, opt_intercept),
        optimized_projection_weights_nemos,
    )

    # check maximization analytical
    # Transition Probability:
    # 1) Compute the lagrange multiplier at the maximum
    #    This is the grad of the loss at the probabilities.
    lagrange_multiplier = -jax.grad(expected_log_likelihood_wrt_transitions)(
        new_transition_prob, xis
    ).mean(
        axis=1
    )  # note that the lagrange mult makes the gradient all the same for each prob.
    # 2) Check that the gradient of the loss is zero
    grad_objective = jax.grad(lagrange_mult_loss)
    (grad_at_transition, grad_at_lagr) = grad_objective(
        (new_transition_prob, lagrange_multiplier),
        xis,
        expected_log_likelihood_wrt_transitions,
    )
    np.testing.assert_array_almost_equal(
        grad_at_transition, np.zeros_like(new_transition_prob)
    )
    np.testing.assert_array_almost_equal(
        grad_at_lagr, np.zeros_like(lagrange_multiplier)
    )
    # Initial probability:
    sum_gammas = np.sum(gammas[np.where(new_sess)[0]], axis=0)
    lagrange_multiplier = -jax.grad(expected_log_likelihood_wrt_initial_prob)(
        new_initial_prob, sum_gammas
    ).mean()  # note that the lagrange mult makes the gradient all the same for each prob.
    grad_objective = jax.grad(lagrange_mult_loss)
    (grad_at_init, grad_at_lagr) = grad_objective(
        (new_initial_prob, lagrange_multiplier),
        sum_gammas,
        expected_log_likelihood_wrt_initial_prob,
    )
    np.testing.assert_array_almost_equal(grad_at_init, np.zeros_like(new_initial_prob))
    np.testing.assert_array_almost_equal(
        grad_at_lagr, np.zeros_like(lagrange_multiplier)
    )


def test_single_state_mstep(single_state_inputs):
    """Single state forward pass posteriors reduces to a GLM."""
    initial_prob, transition_prob, coef, intercept, X, rate, y = single_state_inputs
    obs = PoissonObservations()

    likelihood = jax.vmap(
        lambda x, z: obs.likelihood(x, z, aggregate_sample_scores=lambda w: w),
        in_axes=(None, 1),
        out_axes=1,
    )
    conditionals = likelihood(y, rate)
    new_sess = np.zeros(10)
    new_sess[0] = 1
    alphas, norm = forward_pass(initial_prob, transition_prob, conditionals, new_sess)
    betas = backward_pass(transition_prob, conditionals, norm, new_sess)

    # xis are a sum of the ones over valid entires
    xis = compute_xi(
        alphas,
        betas,
        conditionals,
        norm,
        new_sess,
        transition_prob,
    )

    partial_hmm_negative_log_likelihood, solver = prepare_partial_hmm_nll_single_neuron(
        obs
    )

    (
        optimized_projection_weights_nemos,
        new_initial_prob_nemos,
        new_transition_prob_nemos,
        state,
    ) = run_m_step(
        X,
        y,
        alphas * betas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
    )
    glm = GLM(
        observation_model=obs, solver_name="LBFGS", solver_kwargs={"tol": 10**-13}
    )
    glm.fit(X, y)
    # test that the glm coeff and intercept matches with the m-step output
    np.testing.assert_array_almost_equal(
        glm.coef_, optimized_projection_weights_nemos[0].flatten()
    )
    np.testing.assert_array_almost_equal(
        glm.intercept_, optimized_projection_weights_nemos[1].flatten()
    )

    # test that the transition and initial probabilities are all ones.
    np.testing.assert_array_equal(new_initial_prob_nemos, np.ones_like(initial_prob))
    np.testing.assert_array_equal(
        new_transition_prob_nemos, np.ones_like(new_transition_prob_nemos)
    )

    # check expected shapes
    assert new_transition_prob_nemos.shape == (1, 1)
    assert new_initial_prob_nemos.shape == (1,)
    assert optimized_projection_weights_nemos[0].shape == (2, 1)
    assert optimized_projection_weights_nemos[1].shape == (1,)


def test_maximization_with_prior(generate_data_multi_state):
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state
    )

    obs = PoissonObservations()
    _, solver = prepare_partial_hmm_nll_single_neuron(obs)

    gammas, xis = prepare_gammas_and_xis_for_m_step_single_neuron(
        X, y, initial_prob, transition_prob, (coef, intercept), new_sess, obs
    )

    alphas_transition = np.random.uniform(1, 3, size=transition_prob.shape)
    alphas_init = np.random.uniform(1, 3, size=initial_prob.shape)

    (
        optimized_projection_weights_nemos,
        new_initial_prob,
        new_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=alphas_transition,
        dirichlet_prior_alphas_init_prob=alphas_init,
    )

    lagrange_multiplier = -jax.grad(expected_log_likelihood_wrt_transitions)(
        new_transition_prob, xis, dirichlet_alphas=alphas_transition
    ).mean(
        axis=1
    )  # note that the lagrange mult makes the gradient all the same for each prob.
    # 2) Check that the gradient of the loss is zero
    grad_objective = jax.grad(lagrange_mult_loss)
    (grad_at_transition, grad_at_lagr) = grad_objective(
        (new_transition_prob, lagrange_multiplier),
        xis,
        expected_log_likelihood_wrt_transitions,
        dirichlet_alphas=alphas_transition,
    )
    np.testing.assert_array_almost_equal(
        grad_at_transition, np.zeros_like(new_transition_prob)
    )
    np.testing.assert_array_almost_equal(
        grad_at_lagr, np.zeros_like(lagrange_multiplier)
    )
    # Initial probabilities:
    sum_gammas = np.sum(gammas[np.where(new_sess)[0]], axis=0)
    lagrange_multiplier = -jax.grad(expected_log_likelihood_wrt_initial_prob)(
        new_initial_prob, sum_gammas, dirichlet_alphas=alphas_init
    ).mean()  # note that the lagrange mult makes the gradient all the same for each prob.
    # 2) Check that the gradient of the loss is zero
    grad_objective = jax.grad(lagrange_mult_loss)
    (grad_at_init, grad_at_lagr) = grad_objective(
        (new_initial_prob, lagrange_multiplier),
        sum_gammas,
        expected_log_likelihood_wrt_initial_prob,
        dirichlet_alphas=alphas_init,
    )
    np.testing.assert_array_almost_equal(grad_at_init, np.zeros_like(new_initial_prob))
    np.testing.assert_array_almost_equal(
        grad_at_lagr, np.zeros_like(lagrange_multiplier)
    )


def test_e_and_m_step_for_population(generate_data_multi_state_population):
    """Run E and M step fitting a population."""
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state_population
    )
    obs = PoissonObservations()

    # Wrap likelihood_func to avoid aggregating over samples
    def likelihood_per_sample(x, z):
        return obs.log_likelihood(x, z, aggregate_sample_scores=lambda s: s)

    def negative_log_likelihood_per_sample(x, z):
        return obs._negative_log_likelihood(x, z, aggregate_sample_scores=lambda s: s)

    # Vectorize over the states axis
    state_axes = 2
    likelihood_per_sample = jax.vmap(
        likelihood_per_sample,
        in_axes=(None, state_axes),
        out_axes=state_axes,
    )

    def likelihood(y, rate):
        log_like = likelihood_per_sample(y, rate)
        # Multi-neuron case: sum log-likelihoods across neurons
        log_like = log_like.sum(axis=1)
        return jax.numpy.exp(log_like)

    gammas, xis, _, _, _, _ = forward_backward(
        X,
        y,
        initial_prob,
        transition_prob,
        (coef, intercept),
        likelihood_func=likelihood,
        inverse_link_function=obs.default_inverse_link_function,
        is_new_session=new_sess.astype(bool),
    )

    vmap_nll = jax.vmap(
        negative_log_likelihood_per_sample,
        in_axes=(None, state_axes),
        out_axes=state_axes,
    )

    # solver
    def partial_hmm_negative_log_likelihood(
        weights, design_matrix, observations, posterior_prob
    ):
        return hmm_negative_log_likelihood(
            weights,
            X=design_matrix,
            y=observations,
            posteriors=posterior_prob,
            inverse_link_function=obs.default_inverse_link_function,
            negative_log_likelihood_func=vmap_nll,
        )

    alphas_transition = np.random.uniform(1, 3, size=transition_prob.shape)
    alphas_init = np.random.uniform(1, 3, size=initial_prob.shape)
    solver = LBFGS(partial_hmm_negative_log_likelihood, tol=10**-13)
    (
        optimized_projection_weights_nemos,
        new_initial_prob,
        new_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=alphas_transition,
        dirichlet_prior_alphas_init_prob=alphas_init,
    )


@pytest.mark.parametrize("state_idx", range(5))
def test_m_step_set_alpha_init_to_inf(generate_data_multi_state, state_idx):
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state
    )

    obs = PoissonObservations()

    _, solver = prepare_partial_hmm_nll_single_neuron(obs)

    gammas, xis = prepare_gammas_and_xis_for_m_step_single_neuron(
        X, y, initial_prob, transition_prob, (coef, intercept), new_sess, obs
    )

    alphas_transition = np.random.uniform(1, 3, size=transition_prob.shape)
    alphas_init = np.random.uniform(1, 3, size=initial_prob.shape)
    alphas_init[state_idx] = 10**20

    (
        optimized_projection_weights_nemos,
        new_initial_prob,
        new_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=alphas_transition,
        dirichlet_prior_alphas_init_prob=alphas_init,
    )
    np.testing.assert_array_almost_equal(
        new_initial_prob, np.eye(new_initial_prob.shape[0])[state_idx]
    )


@pytest.mark.parametrize("row, col", itertools.product(range(3), range(3)))
def test_m_step_set_alpha_transition_to_inf(generate_data_multi_state, row, col):
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state
    )

    obs = PoissonObservations()

    _, solver = prepare_partial_hmm_nll_single_neuron(obs)

    gammas, xis = prepare_gammas_and_xis_for_m_step_single_neuron(
        X, y, initial_prob, transition_prob, (coef, intercept), new_sess, obs
    )

    alphas_transition = np.random.uniform(1, 3, size=transition_prob.shape)
    alphas_init = np.random.uniform(1, 3, size=initial_prob.shape)
    alphas_transition[row, col] = 10**20

    (
        optimized_projection_weights_nemos,
        new_initial_prob,
        new_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=alphas_transition,
        dirichlet_prior_alphas_init_prob=alphas_init,
    )
    np.testing.assert_array_almost_equal(
        new_transition_prob[row, :], np.eye(new_initial_prob.shape[0])[col]
    )


def test_m_step_set_alpha_init_to_1(generate_data_multi_state):
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state
    )

    obs = PoissonObservations()
    _, solver = prepare_partial_hmm_nll_single_neuron(obs)

    gammas, xis = prepare_gammas_and_xis_for_m_step_single_neuron(
        X, y, initial_prob, transition_prob, (coef, intercept), new_sess, obs
    )

    alphas_transition = np.random.uniform(1, 3, size=transition_prob.shape)
    alphas_init = np.ones(initial_prob.shape)

    (
        prior_optimized_projection_weights_nemos,
        prior_initial_prob,
        prior_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=alphas_transition,
        dirichlet_prior_alphas_init_prob=alphas_init,
    )
    (
        optimized_projection_weights_nemos,
        no_prior_initial_prob,
        no_prior_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=alphas_transition,
        dirichlet_prior_alphas_init_prob=None,
    )
    np.testing.assert_array_almost_equal(no_prior_initial_prob, prior_initial_prob)
    np.testing.assert_array_almost_equal(
        no_prior_transition_prob, prior_transition_prob
    )
    np.testing.assert_array_almost_equal(
        optimized_projection_weights_nemos[0],
        prior_optimized_projection_weights_nemos[0],
    )
    np.testing.assert_array_almost_equal(
        optimized_projection_weights_nemos[1],
        prior_optimized_projection_weights_nemos[1],
    )


def test_m_step_set_alpha_transition_to_1(generate_data_multi_state):
    new_sess, initial_prob, transition_prob, coef, intercept, X, y = (
        generate_data_multi_state
    )

    obs = PoissonObservations()
    _, solver = prepare_partial_hmm_nll_single_neuron(obs)

    gammas, xis = prepare_gammas_and_xis_for_m_step_single_neuron(
        X, y, initial_prob, transition_prob, (coef, intercept), new_sess, obs
    )
    alphas_transition = np.ones(transition_prob.shape)
    alphas_init = np.random.uniform(1, 3, size=initial_prob.shape)

    (
        prior_optimized_projection_weights_nemos,
        prior_initial_prob,
        prior_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=alphas_transition,
        dirichlet_prior_alphas_init_prob=alphas_init,
    )
    (
        optimized_projection_weights_nemos,
        no_prior_initial_prob,
        no_prior_transition_prob,
        state,
    ) = run_m_step(
        X,
        y,
        gammas,
        xis,
        (np.zeros_like(coef), np.zeros_like(intercept)),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_transition=None,
        dirichlet_prior_alphas_init_prob=alphas_init,
    )
    np.testing.assert_array_almost_equal(no_prior_initial_prob, prior_initial_prob)
    np.testing.assert_array_almost_equal(
        no_prior_transition_prob, prior_transition_prob
    )
    np.testing.assert_array_almost_equal(
        optimized_projection_weights_nemos[0],
        prior_optimized_projection_weights_nemos[0],
    )
    np.testing.assert_array_almost_equal(
        optimized_projection_weights_nemos[1],
        prior_optimized_projection_weights_nemos[1],
    )


@pytest.mark.parametrize(
    "data_name",
    [
        "julia_regression_mstep_flat_prior.npz",  # Uniform priors
        "julia_regression_mstep_good_prior.npz",  # Priors coherent with true parameters
        "julia_regression_mstep_no_prior.npz",  # No prior / uninformative prior (alpha = 1)
    ],
)
def test_run_m_step_regression_priors_simulation(data_name):
    jax.config.update("jax_enable_x64", True)

    # Fetch the data
    data_path = fetch_data(data_name)
    data = np.load(data_path)

    # Design matrix and observed choices
    X, y = data["X"], data["y"]

    # Dirichlet priors
    dirichlet_prior_initial_prob = data[
        "dirichlet_prior_initial_prob"
    ]  # storing alphas (NOT alphas + 1)
    dirichlet_prior_transition_prob = data[
        "dirichlet_prior_transition_prob"
    ]  # storing alphas (NOT alphas + 1)

    # M-step input
    gammas = data["gammas"]
    xis = data["xis"]
    projection_weights = data["projection_weights"]
    intercept, coef = projection_weights[:1], projection_weights[1:]
    new_sess = data["new_sess"]

    # M-step output
    optimized_projection_weights = data["optimized_projection_weights"]
    opt_intercept, opt_coef = (
        optimized_projection_weights[:1],
        optimized_projection_weights[1:],
    )
    new_initial_prob = data["new_initial_prob"]
    new_transition_prob = data["new_transition_prob"]

    # Initialize nemos observation model
    obs = BernoulliObservations()

    # Prepare nll function & solver
    partial_hmm_negative_log_likelihood, solver = prepare_partial_hmm_nll_single_neuron(
        obs
    )

    (
        optimized_projection_weights_nemos,
        new_initial_prob_nemos,
        new_transition_prob_nemos,
        state,
    ) = run_m_step(
        X[:, 1:],  # drop intercept column
        y,
        gammas,
        xis,
        (coef, intercept),
        is_new_session=new_sess.astype(bool),
        solver_run=solver.run,
        dirichlet_prior_alphas_init_prob=dirichlet_prior_initial_prob,
        dirichlet_prior_alphas_transition=dirichlet_prior_transition_prob,
    )

    # NLL with nemos input
    n_ll_nemos = partial_hmm_negative_log_likelihood(
        optimized_projection_weights_nemos,
        X[:, 1:],
        y,
        gammas,
    )

    # NLL with simulation input
    n_ll_original = partial_hmm_negative_log_likelihood(
        (opt_coef, opt_intercept),
        X[:, 1:],
        y,
        gammas,
    )

    # Testing Eq. 13.18 of Bishop
    np.testing.assert_almost_equal(new_initial_prob_nemos, new_initial_prob)
    # Testing Eq. 13.19 of Bishop

    np.testing.assert_almost_equal(
        new_transition_prob_nemos, new_transition_prob, decimal=5
    )  # changed number of decimals

    # Testing output of negative log likelihood
    np.testing.assert_almost_equal(n_ll_original, n_ll_nemos, decimal=10)

    # Testing intercept and optimized projection weights
    jax.tree_util.tree_map(
        lambda x, y: np.testing.assert_almost_equal(x, y, decimal=6),
        (opt_coef, opt_intercept),
        optimized_projection_weights_nemos,
    )



import warnings
from contextlib import nullcontext as does_not_raise

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from nemos.basis.basis import BSplineConv, BSplineEval, RaisedCosineLinearEval
from nemos.identifiability_constraints import (
    _WARN_FLOAT32_MESSAGE,
    _find_drop_column,
    _warn_if_not_float64,
    add_constant,
    apply_identifiability_constraints,
    apply_identifiability_constraints_by_basis_component,
)


@pytest.mark.parametrize(
    "dtype, expected_context, filter_type",
    [
        (
            jnp.int64,
            pytest.warns(
                UserWarning, match="The feature matrix is not of dtype `float64`"
            ),
            "default",
        ),
        (
            jnp.float32,
            pytest.warns(
                UserWarning, match="The feature matrix is not of dtype `float64`"
            ),
            "default",
        ),
        (jnp.float64, does_not_raise(), "error"),
    ],
)
def test_warn_if_not_float64(dtype, expected_context, filter_type):
    """Test warnings raised for float32 matrices."""
    jax.config.update("jax_enable_x64", True)
    matrix = jnp.ones((5, 5), dtype=dtype)
    with expected_context:
        with warnings.catch_warnings():
            warnings.filterwarnings(action=filter_type)
            _warn_if_not_float64(matrix, _WARN_FLOAT32_MESSAGE)


@pytest.mark.parametrize(
    "input_matrix, expected_output",
    [
        (jnp.array([[1, 2], [3, 4]]), jnp.array([[1, 1, 2], [1, 3, 4]])),
        (jnp.array([[0]]), jnp.array([[1, 0]])),
    ],
)
def test_add_constant(input_matrix, expected_output):
    """Test add_constant for adding intercept columns."""
    result = add_constant(input_matrix)
    assert jnp.array_equal(result, expected_output)


@pytest.mark.parametrize(
    "matrix, expected_shape, expected_columns",
    [
        (jnp.array([[1], [0]]), (2, 1), jnp.array([0])),
        (jnp.eye(5), (5, 4), jnp.array([1, 2, 3, 4])),
        (jnp.array([[1, 2], [3, 4]]), (2, 1), jnp.array([1])),
        (jnp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), (3, 1), jnp.array([2])),
        (jnp.array([[1, 0, 0], [0, 0, 0], [0, 0, 1]]), (3, 2), jnp.array([0, 2])),
    ],
)
def test_apply_identifiability_constraints(matrix, expected_shape, expected_columns):
    """Test apply_identifiability_constraints for both full-rank and rank-deficient cases."""
    constrained_x, kept_columns = apply_identifiability_constraints(
        matrix.astype(float), warn_if_float32=False
    )
    assert constrained_x.shape == expected_shape
    assert jnp.array_equal(kept_columns, expected_columns)


@pytest.mark.parametrize(
    "add_intercept, expected_shape", [(True, (2, 0)), (False, (2, 1))]
)
def test_apply_identifiability_constraints_add_constant(add_intercept, expected_shape):
    out, _ = apply_identifiability_constraints(
        np.ones((2, 1)), warn_if_float32=False, add_intercept=add_intercept
    )
    assert out.shape == expected_shape


@pytest.mark.parametrize(
    "basis, input_shape, output_shape, expected_columns",
    [
        (RaisedCosineLinearEval(10, width=4), (50,), (50, 10), jnp.arange(10)),
        (
            BSplineEval(5) + BSplineEval(6),
            (20,),
            (20, 9),
            jnp.array([1, 2, 3, 4, 6, 7, 8, 9, 10]),
        ),
        (
            BSplineConv(5, window_size=10) + BSplineEval(6),
            (20,),
            (20, 10),
            jnp.array([0, 1, 2, 3, 4, 6, 7, 8, 9, 10]),
        ),
        (BSplineEval(5), (10,), (10, 4), jnp.arange(1, 5)),
    ],
)
def test_apply_identifiability_constraints_by_basis_component(
    basis, input_shape, output_shape, expected_columns
):
    """Test constraints applied by basis component."""
    x = basis.compute_features(
        *([np.random.randn(*input_shape)] * basis._n_input_dimensionality)
    )
    constrained_x, kept_columns = apply_identifiability_constraints_by_basis_component(
        basis, x
    )
    assert constrained_x.shape == output_shape
    assert jnp.array_equal(expected_columns, kept_columns)


@pytest.mark.parametrize(
    "matrix, max_drop, preproc, expected_result",
    [
        (
            jnp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
            2,
            add_constant,
            jnp.array([True, True, False]),
        ),
        (
            jnp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
            1,
            add_constant,
            jnp.array([True, False, False]),
        ),
        (jnp.array([[1, 0], [0, 0]]), 1, add_constant, jnp.array([False, True])),
        (jnp.array([[1, 0], [0, 0]]), 0, add_constant, jnp.array([False, False])),
        (np.eye(3, 2), 1, add_constant, jnp.array([False, False])),
        (np.eye(3), 3, add_constant, jnp.array([True, False, False])),
        (
            jnp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
            2,
            lambda x: x,
            jnp.array([True, False, False]),
        ),
        (
            jnp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
            1,
            lambda x: x,
            jnp.array([True, False, False]),
        ),
        (jnp.array([[1, 0], [0, 0]]), 1, lambda x: x, jnp.array([False, True])),
        (jnp.array([[1, 0], [0, 0]]), 0, lambda x: x, jnp.array([False, False])),
        (np.eye(3, 2), 1, lambda x: x, jnp.array([False, False])),
        (np.eye(3), 3, lambda x: x, jnp.array([False, False, False])),
    ],
)
def test_find_drop_column(matrix, max_drop, expected_result, preproc):
    """Test if a column is linearly dependent."""
    jax.config.update("jax_enable_x64", True)
    rank = int(jnp.linalg.matrix_rank(preproc(matrix)))
    result = _find_drop_column(
        matrix.astype(float), rank=rank, max_drop=max_drop, preprocessing_func=preproc
    )
    assert jnp.array_equal(result, expected_result)


@pytest.mark.parametrize(
    "dtype, expected_dtype",
    [
        (jnp.float32, jnp.float32),
        (jnp.float64, jnp.float64),
    ],
)
def test_feature_matrix_dtype(dtype, expected_dtype):
    """Test if the matrix retains its dtype after applying constraints."""
    jax.config.update("jax_enable_x64", True)
    x = np.random.randn(10, 5).astype(dtype)
    constrained_x, _ = apply_identifiability_constraints(x, warn_if_float32=False)
    assert constrained_x.dtype == expected_dtype


@pytest.mark.parametrize(
    "invalid_entries",
    [[np.nan, np.nan], [np.nan, np.inf], [np.inf, np.inf], [np.inf, np.inf]],
)
def test_apply_constraint_with_invalid(invalid_entries):
    """Test if the matrix retains its dtype after applying constraints."""
    jax.config.update("jax_enable_x64", True)
    x = np.random.randn(10, 5)
    # add invalid
    x[:2, 2] = invalid_entries
    # make rank deficient
    x[:, 0] = np.sum(x[:, 1:], axis=1)
    constrained_x, kept_cols = apply_identifiability_constraints(x)
    assert jnp.array_equal(kept_cols, jnp.arange(1, 5))
    assert constrained_x.shape[0] == x.shape[0]
    assert jnp.all(jnp.isnan(constrained_x[:2]))


@pytest.mark.parametrize(
    "invalid_entries",
    [[np.nan, np.nan], [np.nan, np.inf], [np.inf, np.inf], [np.inf, np.inf]],
)
def test_apply_constraint_by_basis_with_invalid(invalid_entries):
    """Test if the matrix retains its dtype after applying constraints."""
    basis = BSplineEval(5)
    x = basis.compute_features(
        np.random.randn(
            10,
        )
    )
    # add invalid
    x[:2, 2] = invalid_entries
    constrained_x, kept_cols = apply_identifiability_constraints(
        x, warn_if_float32=False
    )
    assert jnp.array_equal(kept_cols, jnp.arange(1, 5))
    assert constrained_x.shape[0] == x.shape[0]
    assert jnp.all(jnp.isnan(constrained_x[:2]))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the BFGS solver."""

from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, NamedTuple, Optional, Union
import warnings

import jax
import jax.numpy as jnp
from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _init_stepsize
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _setup_linesearch
from nemos.third_party.jaxopt.jaxopt._src.scipy_wrappers import make_onp_to_jnp
from nemos.third_party.jaxopt.jaxopt._src.scipy_wrappers import pytree_topology_from_example
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_l2_norm
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sub


_dot = partial(jnp.dot, precision=jax.lax.Precision.HIGHEST)
_einsum = partial(jnp.einsum, precision=jax.lax.Precision.HIGHEST)


# Note that BFGS is not meant to be used with high-dimensional problems.
# We support pytrees via flattening.
def pytree_to_flat_array(pytree, dtype):
  """Utility to flatten a pytree."""
  flattened = [jnp.asarray(leaf, dtype).reshape(-1)
               for leaf in jax.tree_util.tree_leaves(pytree)]
  return jnp.concatenate(flattened)


class BfgsState(NamedTuple):
  """Named tuple containing state information."""

  iter_num: int
  value: float
  grad: Any
  stepsize: float
  error: float
  H: jnp.ndarray
  aux: Optional[Any] = None

  num_fun_eval: int = 0
  num_grad_eval: int = 0
  num_linesearch_iter: int = 0


@dataclass(eq=False)
class BFGS(base.IterativeSolver):
  """BFGS solver.

  BFGS is not meant to be used with high-dimensional problems (use LBFGS in this
  case).

  Attributes:
    fun: a smooth function of the form ``fun(x, *args, **kwargs)``.
    value_and_grad: whether ``fun`` just returns the value (False) or both the
      value and gradient (True).
    has_aux: whether ``fun`` outputs auxiliary data or not. If ``has_aux`` is
      False, ``fun`` is expected to be scalar-valued. If ``has_aux`` is True,
      then we have one of the following two cases. If ``value_and_grad`` is
      False, the output should be ``value, aux = fun(...)``. If ``value_and_grad
      == True``, the output should be ``(value, aux), grad = fun(...)``. At each
      iteration of the algorithm, the auxiliary outputs are stored in
      ``state.aux``.
    maxiter: maximum number of proximal gradient descent iterations.
    tol: tolerance of the stopping criterion.
    stepsize: a stepsize to use (if <= 0, use backtracking line search), or a
      callable specifying the **positive** stepsize to use at each iteration.
    linesearch: the type of line search to use: "backtracking" for backtracking
      line search, "zoom" for zoom line search or "hager-zhang" for Hager-Zhang
      line search.
    linesearch_init: strategy for line-search initialization. By default, it
      will use "increase", which will increase the step-size by a factor of
      `increase_factor` at each iteration if the step-size is larger than
      `min_stepsize`, and set it to `max_stepsize` otherwise. Other choices are
      "max", that initializes the step-size to `max_stepsize` at every
      iteration, and "current", that uses the step-size from the previous
      iteration.
    condition: Deprecated. Condition used to select the stepsize when using
      backtracking linesearch.
    maxls: maximum number of iterations to use in the line search.
    decrease_factor: Deprecated. factor by which to decrease the stepsize during
      backtracking line search (default: 0.8).
    increase_factor: factor by which to increase the stepsize during line search
      (default: 1.5).
    max_stepsize: upper bound on stepsize.
    min_stepsize: lower bound on stepsize guess at start of the linesearch run.
    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.
    jit: whether to JIT-compile the optimization loop (default: True).
    unroll: whether to unroll the optimization loop (default: "auto").
    verbose: if set to True or 1 prints the information at each step of 
      the solver, if set to 2, print also the information of the linesearch.

  Reference:
    Jorge Nocedal and Stephen Wright.
    Numerical Optimization, second edition.
    Algorithm 6.1 (page 140).
  """

  fun: Callable
  value_and_grad: bool = False
  has_aux: bool = False

  maxiter: int = 500
  # FIXME: should depend on whether float32 or float64 is used.
  # Tests should pass in float64 without modifying tol
  tol: float = 1e-3

  stepsize: Union[float, Callable] = 0.0
  linesearch: str = "zoom"
  linesearch_init: str = "increase"
  condition: Any = None  # deprecated in v0.8
  maxls: int = 30
  decrease_factor: Any = None  # deprecated in v0.8
  increase_factor: float = 1.5
  max_stepsize: float = 1.0
  # FIXME: should depend on whether float32 or float64 is used.
  min_stepsize: float = 1e-6

  implicit_diff: bool = True
  implicit_diff_solve: Optional[Callable] = None

  jit: bool = True
  unroll: base.AutoOrBoolean = "auto"

  verbose: Union[bool, int] = False

  def init_state(self,
                 init_params: Any,
                 *args,
                 **kwargs) -> BfgsState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      state
    """
    (value, aux), grad = self._value_and_grad_with_aux(init_params, *args, **kwargs)

    dtype = tree_single_dtype(init_params)
    flat_init_params = pytree_to_flat_array(init_params, dtype)

    return BfgsState(iter_num=jnp.asarray(0),
                     value=value,
                     grad=grad,
                     stepsize=jnp.asarray(self.max_stepsize, dtype=dtype),
                     error=jnp.asarray(jnp.inf, dtype=dtype),
                     H=jnp.eye(len(flat_init_params), dtype=dtype),
                     aux=aux,
                     num_fun_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
                     num_grad_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
                     num_linesearch_iter=jnp.asarray(0, base.NUM_EVAL_DTYPE)
                     )

  def update(self,
             params: Any,
             state: BfgsState,
             *args,
             **kwargs) -> base.OptStep:
    """Performs one iteration of BFGS.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      (params, state)
    """
    value, grad = state.value, state.grad
    pytree_topology = pytree_topology_from_example(params)
    flat_array_to_pytree = make_onp_to_jnp(pytree_topology)
    dtype = tree_single_dtype(params)
    flat_grad = pytree_to_flat_array(grad, dtype)

    descent_direction = flat_array_to_pytree(-_dot(state.H, flat_grad))

    use_linesearch = not isinstance(self.stepsize, Callable) and self.stepsize <= 0

    if use_linesearch:
      init_stepsize = _init_stepsize(
          self.linesearch_init,
          self.max_stepsize,
          self.min_stepsize,
          self.increase_factor,
          state.stepsize,
      )
      new_stepsize, ls_state = self.run_ls(
          init_stepsize,
          params,
          value,
          grad,
          descent_direction,
          fun_args=args,
          fun_kwargs=kwargs,
      )
      new_params = ls_state.params
      new_value = ls_state.value
      new_grad = ls_state.grad
      new_aux = ls_state.aux
      new_num_linesearch_iter = state.num_linesearch_iter + ls_state.iter_num
      new_num_grad_eval = state.num_grad_eval + ls_state.num_grad_eval
      new_num_fun_eval = state.num_fun_eval + ls_state.num_fun_eval
    else:
      if isinstance(self.stepsize, Callable):
        new_stepsize = self.stepsize(state.iter_num)
      else:
        new_stepsize = self.stepsize

      new_params = tree_add_scalar_mul(params, new_stepsize, descent_direction)
      (new_value, new_aux), new_grad = self._value_and_grad_with_aux(new_params, *args, **kwargs)
      new_num_grad_eval = state.num_grad_eval + 1
      new_num_fun_eval = state.num_fun_eval + 1
      new_num_linesearch_iter = state.num_linesearch_iter

    s = tree_sub(new_params, params)
    y = tree_sub(new_grad, grad)
    flat_s = pytree_to_flat_array(s, dtype)
    flat_y = pytree_to_flat_array(y, dtype)
    rho = jnp.reciprocal(_dot(flat_y, flat_s))

    sy = jnp.outer(flat_s, flat_y)
    ss = jnp.outer(flat_s, flat_s)
    w = jnp.eye(len(flat_grad), dtype=rho.dtype) - rho * sy
    new_H = _einsum('ij,jk,lk', w, state.H, w) + rho * ss
    new_H = jnp.where(jnp.isfinite(rho), new_H, state.H)

    error = tree_l2_norm(new_grad)
    new_state = BfgsState(iter_num=state.iter_num + 1,
                          value=new_value,
                          grad=new_grad,
                          stepsize=jnp.asarray(new_stepsize),
                          error=jnp.asarray(error, dtype=dtype),
                          H=new_H,
                          aux=new_aux,
                          num_grad_eval=new_num_grad_eval,
                          num_fun_eval=new_num_fun_eval,
                          num_linesearch_iter=new_num_linesearch_iter)

    if self.verbose:
      self.log_info(
          new_state,
          error_name="Gradient Norm",
          additional_info={
              "Objective Value": new_value,
              "Stepsize": new_stepsize,
              "Number Linesearch Iterations": 
              new_state.num_linesearch_iter - state.num_linesearch_iter
          }
      )
    return base.OptStep(params=new_params, state=new_state)

  def optimality_fun(self, params, *args, **kwargs):
    """Optimality function mapping compatible with ``@custom_root``."""
    return self._value_and_grad_fun(params, *args, **kwargs)[1]

  def _value_and_grad_fun(self, params, *args, **kwargs):
    (value, _), grad = self._value_and_grad_with_aux(params, *args, **kwargs)
    return value, grad

  def __post_init__(self):
    super().__post_init__()

    _fun_with_aux, _, self._value_and_grad_with_aux = \
      base._make_funs_with_aux(fun=self.fun,
                               value_and_grad=self.value_and_grad,
                               has_aux=self.has_aux)

    self.reference_signature = self.fun
    unroll = self._get_unroll_option()
    self.linesearch_solver = _setup_linesearch(
        linesearch=self.linesearch,
        fun=_fun_with_aux,
        value_and_grad=self._value_and_grad_with_aux,
        has_aux=True,
        maxlsiter=self.maxls,
        max_stepsize=self.max_stepsize,
        jit=self.jit,
        unroll=unroll,
        verbose=max(int(self.verbose)-1, 0)
    )
    self.run_ls = self.linesearch_solver.run

    # FIXME: to remove in future releases
    if self.condition is not None:
      warnings.warn("Argument condition is deprecated", DeprecationWarning)
    if self.decrease_factor is not None:
      warnings.warn(
          "Argument decrease_factor is deprecated", DeprecationWarning
      )



"""Basis module."""

from ._basis import AdditiveBasis, MultiplicativeBasis
from ._composition_utils import __PUBLIC_BASES__
from ._custom_basis import CustomBasis
from ._transformer_basis import TransformerBasis
from .basis import (
    BSplineConv,
    BSplineEval,
    CyclicBSplineConv,
    CyclicBSplineEval,
    FourierEval,
    HistoryConv,
    IdentityEval,
    MSplineConv,
    MSplineEval,
    OrthExponentialConv,
    OrthExponentialEval,
    RaisedCosineLinearConv,
    RaisedCosineLinearEval,
    RaisedCosineLogConv,
    RaisedCosineLogEval,
)

__all__ = __PUBLIC_BASES__


def __dir__():
    return __all__



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Isotonic Regression."""

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from jax.test_util import check_grads
from nemos.third_party.jaxopt.jaxopt.isotonic import isotonic_l2_pav
from nemos.third_party.jaxopt.jaxopt._src import test_util
from sklearn import isotonic


class IsotonicPavTest(test_util.JaxoptTestCase):
  """Tests for PAV in JAX."""

  def test_output_shape_and_dtype(self, n=10):
    """Verifies the shapes and dtypes of output."""
    y = jax.random.normal(jax.random.PRNGKey(0), (n,))
    output = isotonic_l2_pav(y)
    self.assertEqual(output.shape, y.shape)
    self.assertEqual(output.dtype, y.dtype)

  @parameterized.product(increasing=[True, False])
  def test_compare_with_sklearn(self, increasing,  n=10):
    """Compares the output with the one of sklearn."""
    y = jax.random.normal(jax.random.PRNGKey(0), (n,))
    output = isotonic_l2_pav(y, increasing=increasing)
    output_sklearn = jnp.array(isotonic.isotonic_regression(y, increasing=increasing))
    self.assertArraysAllClose(output, output_sklearn)
    y_sort = y.sort()
    y_min = y_sort[2]
    y_max = y_sort[n-5]
    output = isotonic_l2_pav(y, y_min=y_min, y_max=y_max, increasing=increasing)
    output_sklearn = jnp.array(isotonic.isotonic_regression(y, y_min=y_min.item(),
     y_max=y_max.item(), increasing=increasing))
    self.assertArraysAllClose(output, output_sklearn)

  @parameterized.product(increasing=[True, False])
  def test_gradient(self, increasing, n=10):
    """Checks the gradient with finite differences."""
    # Absolute error of test fails for large values of y.
    y = 0.1*jax.random.normal(jax.random.PRNGKey(0), (n,))

    def loss(y):
      return (isotonic_l2_pav(y**3, increasing=increasing)
       + isotonic_l2_pav(y, increasing=increasing) ** 2).mean()

    check_grads(loss, (y,), order=2)

  def test_gradient_min_max(self, n=10):
    """Checks the gradient with finite differences."""
    y = jax.random.normal(jax.random.PRNGKey(0), (n,))
    y_sort = y.sort()
    y_min = y_sort[2]
    y_max = y_sort[n-5]
    def loss(y):
      return (isotonic_l2_pav(y**3, y_min=y_min, y_max=y_max)
       + isotonic_l2_pav(y, y_min=y_min, y_max=y_max) ** 2).mean()

    check_grads(loss, (y,), order=2)

  def test_vmap(self, n_features=10, n_batches=16):
    """Verifies vmap."""
    y = jax.random.normal(jax.random.PRNGKey(0), (n_batches, n_features))
    isotonic_l2_pav_vmap = jax.vmap(isotonic_l2_pav)
    for i in range(n_batches):
      self.assertArraysAllClose(isotonic_l2_pav_vmap(y)[i], isotonic_l2_pav(y[i]))

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from absl.testing import absltest

import jax.numpy as jnp
from jax.test_util import check_grads
import optax

from nemos.third_party.jaxopt.jaxopt import objective

from nemos.third_party.jaxopt.jaxopt import prox
from nemos.third_party.jaxopt.jaxopt._src import test_util

from nemos.third_party.jaxopt.jaxopt import AndersonWrapper
from nemos.third_party.jaxopt.jaxopt import BlockCoordinateDescent
from nemos.third_party.jaxopt.jaxopt import OptaxSolver
from nemos.third_party.jaxopt.jaxopt import PolyakSGD
from nemos.third_party.jaxopt.jaxopt import ProximalGradient

from sklearn import datasets


class AndersonWrapperTest(test_util.JaxoptTestCase):

  def test_proximal_gradient_wrapper(self):
    """Baseline test on simple optimizer."""
    X, y = datasets.make_regression(n_samples=100, n_features=20, random_state=0)
    fun = objective.least_squares
    lam = 10.0
    data = (X, y)
    w_init = jnp.zeros(X.shape[1])
    tol = 1e-3
    maxiter = 1000
    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso, maxiter=maxiter, tol=tol,
                          acceleration=False)
    aw = AndersonWrapper(pg, history_size=15)
    aw_params, awpg_info = aw.run(w_init, hyperparams_prox=lam, data=data)
    self.assertLess(awpg_info.error, tol)

  def test_mixing_frequency_polyak(self):
    """Test mixing_frequency by accelerating PolyakSGD."""
    X, y = datasets.make_classification(n_samples=10, n_features=5, n_classes=3,
                                        n_informative=3, random_state=0)
    data = (X, y)
    l2reg = 100.0
    # fun(params, data)
    fun = objective.l2_multiclass_logreg_with_intercept
    n_classes = len(jnp.unique(y))

    W_init = jnp.zeros((X.shape[1], n_classes))
    b_init = jnp.zeros(n_classes)
    pytree_init = (W_init, b_init)

    opt = PolyakSGD(fun=fun, max_stepsize=0.01, tol=0.05, momentum=False)
    history_size = 5
    aw = AndersonWrapper(opt, history_size=history_size, mixing_frequency=1)
    aw_params, aw_state = aw.run(pytree_init, l2reg=l2reg, data=data)
    self.assertLess(aw_state.error, 0.05)

  def test_optax_restart(self):
    """Test Optax optimizer."""
    X, y = datasets.make_classification(n_samples=100, n_features=20, n_classes=3,
                                        n_informative=3, random_state=0)
    data = (X, y)
    l2reg = 100.0
    # fun(params, data)
    fun = objective.l2_multiclass_logreg_with_intercept
    n_classes = len(jnp.unique(y))

    W_init = jnp.zeros((X.shape[1], n_classes))
    b_init = jnp.zeros(n_classes)
    pytree_init = (W_init, b_init)

    tol = 1e-2
    opt = OptaxSolver(opt=optax.sgd(1e-2, momentum=0.8), fun=fun, maxiter=1000, tol=0)
    aw = AndersonWrapper(opt, history_size=3, ridge=1e-3)
    params, infos = aw.run(pytree_init, l2reg=l2reg, data=data)

    # Check optimality conditions.
    error = opt.l2_optimality_error(params, l2reg=l2reg, data=data)
    self.assertLessEqual(error, tol)

  def test_block_cd_restart(self):
    """Accelerate Block CD."""
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)

    # Setup parameters.
    fun = objective.least_squares  # fun(params, data)
    l1reg = 10.0
    data = (X, y)

    # Initialize.
    w_init = jnp.zeros(X.shape[1])
    tol = 5e-4
    maxiter = 100
    bcd = BlockCoordinateDescent(fun=fun, block_prox=prox.prox_lasso, tol=tol, maxiter=maxiter)
    aw = AndersonWrapper(bcd, history_size=3, ridge=1e-5)
    params, state = aw.run(init_params=w_init, hyperparams_prox=l1reg, data=data)

    # Check optimality conditions.
    self.assertLess(state.error, tol)

  def test_wrapper_grad(self):
    """Test gradient of wrapper."""
    data_train = datasets.make_regression(n_samples=100, n_features=3, random_state=0)
    fun = objective.least_squares
    lam = 10.0
    w_init = jnp.zeros(data_train[0].shape[1])
    tol = 1e-5
    maxiter = 1000  # large number of updates
    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso, maxiter=maxiter, tol=tol,
                          acceleration=False)
    aw = AndersonWrapper(pg, history_size=5)
    data_val = datasets.make_regression(n_samples=100, n_features=3, random_state=0)

    def solve_run(lam):
      aw_params = aw.run(w_init, lam, data_train).params
      loss = fun(aw_params, data=data_val)
      return loss

    check_grads(solve_run, args=(lam,), order=1, modes=['rev'], eps=2e-2)

    def solve_run(lam):
      aw_params = aw.run(w_init, hyperparams_prox=lam, data=data_train).params
      loss = fun(aw_params, data=data_val)
      return loss

    check_grads(solve_run, args=(lam,), order=1, modes=['rev'], eps=2e-2)

if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # config.update("jax_enable_x64", True)
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

import nemos.third_party.jaxopt.jaxopt as jaxopt
from nemos.third_party.jaxopt.jaxopt._src import test_util


class ImportTest(test_util.JaxoptTestCase):
    def test_implicit_diff(self):
        jaxopt.implicit_diff.root_vjp
        from nemos.third_party.jaxopt.jaxopt.implicit_diff import root_vjp

    def test_isotonic(self):
        jaxopt.isotonic.isotonic_l2_pav
        from nemos.third_party.jaxopt.jaxopt.isotonic import isotonic_l2_pav

    def test_prox(self):
        jaxopt.prox.prox_none
        from nemos.third_party.jaxopt.jaxopt.prox import prox_none

    def test_projection(self):
        jaxopt.projection.projection_simplex
        from nemos.third_party.jaxopt.jaxopt.projection import projection_simplex

    def test_tree_util(self):
        from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot

    def test_linear_solve(self):
        from nemos.third_party.jaxopt.jaxopt.linear_solve import solve_lu

    def test_base(self):
        from nemos.third_party.jaxopt.jaxopt.base import LinearOperator

    def test_perturbations(self):
        from nemos.third_party.jaxopt.jaxopt.perturbations import make_perturbed_argmax

    def test_loss(self):
        jaxopt.loss.binary_logistic_loss
        from nemos.third_party.jaxopt.jaxopt.loss import binary_logistic_loss

    def test_objective(self):
        jaxopt.objective.least_squares
        from nemos.third_party.jaxopt.jaxopt.objective import least_squares

    def test_loop(self):
        from nemos.third_party.jaxopt.jaxopt.loop import while_loop


if __name__ == "__main__":
    absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import prox
from nemos.third_party.jaxopt.jaxopt import ProximalGradient
from nemos.third_party.jaxopt.jaxopt._src import test_util

from sklearn import datasets


def make_stepsize_schedule(max_stepsize, n_steps, power=1.0) -> Callable:
  def stepsize_schedule(t: int) -> float:
    true_fn = lambda t: 1.0
    false_fn = lambda t: (n_steps / t) ** power
    decay_factor = jax.lax.cond(t <= n_steps, true_fn, false_fn, t)
    return decay_factor * max_stepsize
  return stepsize_schedule


N_CALLS = 0

class ProximalGradientTest(test_util.JaxoptTestCase):

  @parameterized.product(acceleration=[True, False])
  def test_lasso_manual_loop(self, acceleration):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    fun = objective.least_squares  # fun(params, data)
    lam = 10.0
    data = (X, y)

    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso,
                          acceleration=acceleration)
    params = jnp.zeros(X.shape[1])
    state = pg.init_state(params, hyperparams_prox=lam, data=data)
    for _ in range(10):
      params, state = pg.update(params, state, hyperparams_prox=lam, data=data)

    # Check optimality conditions.
    self.assertLess(state.error, 2e-2)

  @parameterized.product(acceleration=[True, False])
  def test_lasso(self, acceleration=True):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    fun = objective.least_squares
    lam = 10.0
    data = (X, y)

    w_init = jnp.zeros(X.shape[1])
    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso, maxiter=200, tol=1e-3,
                          acceleration=acceleration)
    w_fit, info = pg.run(w_init, hyperparams_prox=lam, data=data)

    # Check optimality conditions.
    self.assertLess(info.error, 1e-3)

    # Compare against sklearn.
    w_skl = test_util.lasso_skl(X, y, lam)
    self.assertArraysAllClose(w_fit, w_skl, atol=1e-2)

  def test_lasso_implicit_diff(self):
    """Test implicit differentiation of a single lambda parameter."""
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    lam = 10.0
    data = (X, y)

    fun = objective.least_squares
    jac_num = test_util.lasso_skl_jac(X, y, lam)
    w_skl = test_util.lasso_skl(X, y, lam)

    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso, tol=1e-3, maxiter=200,
                          acceleration=True, implicit_diff=True)

    def wrapper(hyperparams_prox):
      return pg.run(w_skl, hyperparams_prox, data).params

    jac_prox = jax.jacrev(wrapper)(lam)
    self.assertArraysAllClose(jac_num, jac_prox, atol=1e-3)

  def test_lasso_implicit_diff_with_kwargs(self):
    """Same as above but with kwargs."""
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    lam = 10.0
    data = (X, y)

    fun = objective.least_squares
    jac_num = test_util.lasso_skl_jac(X, y, lam)
    w_skl = test_util.lasso_skl(X, y, lam)

    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso, tol=1e-3, maxiter=200,
                          acceleration=True, implicit_diff=True)

    def wrapper(hyperparams_prox):
      return pg.run(w_skl, hyperparams_prox=hyperparams_prox, data=data).params

    jac_prox = jax.jacrev(wrapper)(lam)
    self.assertArraysAllClose(jac_num, jac_prox, atol=1e-3)

  def test_stepsize_schedule(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    fun = objective.least_squares  # fun(params, data)
    data = (X, y)

    schedule = make_stepsize_schedule(max_stepsize=1., n_steps=80, power=1.0)

    pg = ProximalGradient(fun=fun, stepsize=schedule,
                          prox=prox.prox_none,
                          acceleration=False)
    params = jnp.zeros(X.shape[1])
    _, state = pg.run(params, hyperparams_prox=None, data=data)
    self.assertLess(state.error, 1e-3)

  @parameterized.product(n_iter=[10])
  def test_n_calls(self, n_iter):
    """Test whether the number of function calls
    is equal to the number of iterations + 1 in the
    no linesearch case, where the complexity is linear."""
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    orig_fun = objective.least_squares
    def fun(params, data):
      global N_CALLS
      N_CALLS += 1
      return orig_fun(params, data)
    lam = 10.0
    data = (X, y)

    w_init = jnp.zeros(X.shape[1])
    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso, maxiter=n_iter, tol=1e-10, stepsize=1e-5,
                          acceleration=True, jit=False)
    w_fit, info = pg.run(w_init, hyperparams_prox=lam, data=data)

    self.assertEqual(N_CALLS, n_iter)


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt._src.lbfgs import LBFGS
from nemos.third_party.jaxopt.jaxopt import OSQP

import numpy as onp


class ProjectionTest(test_util.JaxoptTestCase):

  def test_projection_non_negative(self):
    x = jnp.array([-1.0, 2.0, 3.0])
    expected = jnp.array([0, 2.0, 3.0])
    self.assertArraysEqual(projection.projection_non_negative(x), expected)
    self.assertArraysEqual(projection.projection_non_negative((x, x)),
                           (expected, expected))
    # test with nested pytree
    tree_x = (-1.0, {"k1": 1.0, "k2": (1.0, 1.0)}, 1.0)
    tree_expected = (0.0, {"k1": 1.0, "k2": (1.0, 1.0)}, 1.0)
    self.assertAllClose(projection.projection_non_negative(tree_x),
                        tree_expected)

  def test_projection_box(self):
    x = jnp.array([-1.0, 2.0, 3.0])
    expected = jnp.array([0, 2.0, 2.0])

    L, U = 0.0, 2.0

    # lower and upper values are scalars.
    self.assertArraysEqual(projection.projection_box(x, (L, U)), expected)
    params = ((L, L), (U, U))
    self.assertArraysEqual(projection.projection_box((x, x), params),
                           (expected, expected))

    # lower and upper are pytrees. Note: Does not work when U and L are scalars!
    tree_x = (-1.0, {"k1": 2.0, "k2": (2.0, 3.0)}, 3.0)
    tree_expected = (0.0, {"k1": 2.0, "k2": (2.0, 2.0)}, 2.0)
    U_tree = (2.0, {"k1": 2.0, "k2": (2.0, 2.0)}, 2.0)
    L_tree = (0.0, {"k1": 0.0, "k2": (0.0, 0.0)}, 0.0)
    self.assertArraysEqual(projection.projection_box(tree_x, [L_tree, U_tree]),
                           tree_expected)

    # lower and upper values are arrays.
    params = (jnp.ones(len(x)) * L, jnp.ones(len(x)) * U)
    self.assertArraysEqual(projection.projection_box(x, params), expected)
    params = ((params[0], params[0]), (params[1], params[1]))
    self.assertArraysEqual(projection.projection_box((x, x), params),
                           (expected, expected))

  def test_projection_hypercube(self):
    x = jnp.array([-1.0, 2.0, 0.5])

    expected = jnp.array([0, 1.0, 0.5])
    self.assertArraysEqual(projection.projection_hypercube(x), expected)

    expected = jnp.array([0, 0.8, 0.5])
    self.assertArraysEqual(projection.projection_hypercube(x, 0.8), expected)


  def test_projection_simplex(self):
    jax.config.update("jax_enable_x64", True)
    rng = onp.random.RandomState(0)

    for scale in [1.0, 1e9]:
      for _ in range(10):
        x = scale * rng.randn(50).astype(onp.float64)

        p = projection.projection_simplex(x)
        self.assertAllClose(jnp.sum(p), 1.0)
        self.assertTrue(jnp.all(0 <= p))
        self.assertTrue(jnp.all(p <= 1))

        p = projection.projection_simplex(x, 0.8)
        self.assertAllClose(jnp.sum(p), 0.8)
        self.assertTrue(jnp.all(0 <= p))
        self.assertTrue(jnp.all(p <= 0.8))

    # Check that -inf is supported.
    p = projection.projection_simplex(jnp.array([0.0, 0.0, -jnp.inf]))
    self.assertAllClose(p, jnp.array([0.5, 0.5, 0.0]))

  def test_projection_simplex_jacobian(self):
    jax.config.update("jax_enable_x64", True)
    rng = onp.random.RandomState(0)

    for scale in [1.0, 1e9]:
      x = scale * rng.rand(5).astype(onp.float64)
      v = scale * rng.randn(5).astype(onp.float64)

      J_rev = jax.jacrev(projection.projection_simplex)(x)
      J_fwd = jax.jacfwd(projection.projection_simplex)(x)

      # Check against theoretical expression.
      p = projection.projection_simplex(x)
      support = (p > 0).astype(jnp.int32)
      cardinality = jnp.count_nonzero(support)
      J_true = jnp.diag(support) - jnp.outer(support, support) / cardinality
      self.assertArraysAllClose(J_true, J_fwd)
      self.assertArraysAllClose(J_true, J_rev)

      # Check vector Jacobian product.
      vJ, = jax.vjp(projection.projection_simplex, x)[1](v)
      self.assertArraysAllClose(vJ, jnp.dot(v, J_true))

      # Check vector Jacobian product.
      Jv = jax.jvp(projection.projection_simplex, (x,), (v,))[1]
      self.assertArraysAllClose(Jv, jnp.dot(J_true, v))

  def test_projection_simplex_vmap(self):
    rng = onp.random.RandomState(0)
    X = rng.randn(3, 50).astype(onp.float64)

    # Check with default s=1.0.
    P = jax.vmap(projection.projection_simplex)(X)
    self.assertArraysAllClose(jnp.sum(P, axis=1), jnp.ones(len(X)))
    self.assertTrue(jnp.all(0 <= P))
    self.assertTrue(jnp.all(P <= 1))

    # Check with s=0.8.
    P = jax.vmap(projection.projection_simplex)(X, 0.8 * jnp.ones(len(X)))
    self.assertArraysAllClose(jnp.sum(P, axis=1), 0.8 * jnp.ones(len(X)))
    self.assertTrue(jnp.all(0 <= P))
    self.assertTrue(jnp.all(P <= 0.8))

  def test_projection_simplex_vmap_diff(self):
    proj = jax.vmap(projection.projection_simplex)

    def fun(X):
      return jnp.sum(proj(X) ** 2)

    rng = onp.random.RandomState(0)
    X = rng.rand(4, 5).astype(onp.float32)
    U = rng.rand(4, 5)
    U /= onp.sqrt(onp.sum(U ** 2))
    U = U.astype(onp.float32)

    eps = 1e-3
    dir_deriv_num = (fun(X + eps * U) - fun(X - eps * U)) / (2 * eps)
    dir_deriv = jnp.vdot(jax.grad(fun)(X), U)
    self.assertAllClose(dir_deriv, dir_deriv_num, atol=1e-3)

  def test_projection_l1_sphere(self):
    rng = onp.random.RandomState(0)

    x = rng.randn(50).astype(onp.float32)

    p = projection.projection_l1_sphere(x)
    self.assertAllClose(jnp.sum(jnp.abs(p)), 1.0, check_dtypes=False)

    radius = 3.21
    p = projection.projection_l1_sphere(x, radius)
    self.assertAllClose(jnp.sum(jnp.abs(p)), radius, check_dtypes=False)

  def _check_projection_ball(self, ball_fun, sphere_fun, norm, check_pytree):
    rng = onp.random.RandomState(0)

    x = rng.randn(50).astype(onp.float32)
    norm_value = jnp.linalg.norm(x, ord=norm)

    # x is already in the ball
    big_radius = norm_value * 2
    p = ball_fun(x, big_radius)
    self.assertArraysAllClose(x, p, check_dtypes=False)
    J = jax.jacrev(ball_fun)(x, big_radius)
    self.assertArraysAllClose(J, jnp.eye(len(x)))
    J = jax.jacrev(ball_fun, argnums=1)(x, big_radius)
    self.assertArraysAllClose(J, jnp.zeros_like(x), check_dtypes=False)

    # x is at the boundary of the ball
    p = ball_fun(x, norm_value)
    self.assertArraysAllClose(x, p, check_dtypes=False)
    J = jax.jacrev(ball_fun)(x, norm_value)
    self.assertArraysAllClose(J, jnp.eye(len(x)), check_dtypes=False)
    J = jax.jacrev(ball_fun, argnums=1)(x, norm_value)
    self.assertArraysAllClose(J, jnp.zeros_like(x), check_dtypes=False)

    # x is outside of the ball
    small_radius = norm_value / 2
    p = ball_fun(x, small_radius)
    self.assertAllClose(jnp.linalg.norm(p, ord=norm), small_radius, check_dtypes=False)
    J = jax.jacrev(ball_fun)(x, small_radius)
    J2 = jax.jacrev(sphere_fun)(x, small_radius)
    self.assertArraysAllClose(J, J2, check_dtypes=False)
    J = jax.jacrev(ball_fun, argnums=1)(x, small_radius)
    J2 = jax.jacrev(sphere_fun, argnums=1)(x, small_radius)
    self.assertArraysAllClose(J, J2, check_dtypes=False)

    if check_pytree:
      # x is a pytree
      pytree = (x[:25], x[25:])
      p = ball_fun(pytree, norm_value)
      self.assertArraysAllClose(x[:25], p[0], check_dtypes=False)
      self.assertArraysAllClose(x[25:], p[1], check_dtypes=False)

  def test_projection_l1_ball(self):
    self._check_projection_ball(projection.projection_l1_ball,
                                projection.projection_l1_sphere,
                                norm=1, check_pytree=False)

  def test_projection_l2_sphere(self):
    rng = onp.random.RandomState(0)

    x = rng.randn(50).astype(onp.float32)

    p = projection.projection_l2_sphere(x)
    self.assertAllClose(jnp.sum(p ** 2), 1.0, check_dtypes=False)

    radius = 3.21
    p = projection.projection_l2_sphere(x, radius)
    self.assertAllClose(jnp.sqrt(jnp.sum(p ** 2)), radius, check_dtypes=False)

  def test_projection_l2_ball(self):
    self._check_projection_ball(projection.projection_l2_ball,
                                projection.projection_l2_sphere,
                                norm=2, check_pytree=True)

  def test_projection_linf_ball(self):
    rng = onp.random.RandomState(0)

    x = rng.randn(50).astype(onp.float32)
    linf_norm = jnp.max(jnp.abs(x))

    p = projection.projection_linf_ball(x, linf_norm)
    self.assertArraysAllClose(x, p)

    p = projection.projection_linf_ball(x, linf_norm * 0.5)
    self.assertAllClose(jnp.max(jnp.abs(p)), linf_norm * 0.5)

  def test_projection_hyperplane(self):
    rng = onp.random.RandomState(0)

    x = rng.randn(50).astype(onp.float32)
    a = rng.randn(50).astype(onp.float32)
    b = 1.0

    p = projection.projection_hyperplane(x, (a, b))
    self.assertAllClose(jnp.dot(a, p), b, check_dtypes=False)

    tree_x = (1.0, {"k1": 2.0, "k2": (2.0, 1.0)}, 2.0)
    tree_a = (1.0, {"k1": 1.0, "k2": (1.0, 0.0)}, 1.0)
    p = projection.projection_hyperplane(tree_x, (tree_a, b))
    expected_p = (-0.5, {"k1": 0.5, "k2": (0.5, 1.0)}, 0.5)
    self.assertAllClose(expected_p, p, check_dtypes=False)

  def test_projection_halfspace(self):
    # For b very large, halfplane projection of x is x.
    rng = onp.random.RandomState(0)
    tree_x = (rng.randn(1), {"k1": rng.randn(1), "k2": (rng.randn(1), rng.randn(1))}, rng.randn(1))
    tree_a = (rng.randn(1), {"k1": rng.randn(1), "k2": (rng.randn(1), rng.randn(1))}, rng.randn(1))
    b = 10000
    p = projection.projection_halfspace(tree_x, (tree_a, b))
    self.assertAllClose(tree_x, p)

    # For b very negative, the halfplane projection of x is on the hyperplane <a,x> =b.
    b = -10000
    p = projection.projection_halfspace(tree_x, (tree_a, b))
    p_hyper = projection.projection_hyperplane(tree_x, (tree_a, b))
    self.assertAllClose(p_hyper, p)

  def test_projection_affine_set(self):
    rng = onp.random.RandomState(0)

    x = rng.randn(5).astype(onp.float32)
    A = rng.randn(3, 5).astype(onp.float32)
    b = rng.randn(3).astype(onp.float32)
    params = (A, b)

    p = projection.projection_affine_set(x, params)
    self.assertAllClose(jnp.dot(A, p), b, atol=1e-3, rtol=1e-3)

    jac_fun = jax.jacrev(projection.projection_affine_set, argnums=0)
    jac_x = jac_fun(x, params)
    self.assertEqual(jac_x.shape, (5, 5))

    jac_fun = jax.jacrev(projection.projection_affine_set, argnums=1)
    jac_A, jac_b = jac_fun(x, params)
    self.assertEqual(jac_A.shape, (5, 3, 5))
    self.assertEqual(jac_b.shape, (5, 3))

  def test_projection_polyhedron(self):
    jax.config.update("jax_enable_x64", True)
    rng = onp.random.RandomState(0)

    x = rng.randn(5).astype(onp.float64)
    A = rng.randn(3, 5).astype(onp.float64)
    b = rng.randn(3).astype(onp.float64)
    G = rng.randn(2, 5).astype(onp.float64)
    h = rng.randn(2).astype(onp.float64)
    params = (A, b, G, h)

    p = projection.projection_polyhedron(x, params)
    self.assertAllClose(jnp.dot(A, p), b, atol=1e-3, rtol=1e-3)

    jac_fun = jax.jacrev(projection.projection_polyhedron, argnums=0)
    jac_x = jac_fun(x, params)
    self.assertEqual(jac_x.shape, (5, 5))

    jac_fun = jax.jacrev(projection.projection_polyhedron, argnums=1)
    jac_A, jac_b, jac_G, jac_h = jac_fun(x, params)
    self.assertEqual(jac_A.shape, (5, 3, 5))
    self.assertEqual(jac_b.shape, (5, 3))
    self.assertEqual(jac_G.shape, (5, 2, 5))
    self.assertEqual(jac_h.shape, (5, 2))

  def test_projection_polyhedron_infeasible(self):
    def proj(x):
      A = jnp.array([[1.0, 1.0],[0.0, 1.0],[-1.0, 1.0]])
      b = jnp.array([1.0, 1.0, -1.0])
      G = jnp.array([[-1.0, 0.0], [0.0, -1.0]])
      h = jnp.array([0.0, 0.0])
      return projection.projection_polyhedron(x, hyperparams=(A, b, G, h))

    self.assertRaises(ValueError, proj, jnp.zeros(2))

  def test_projection_box_section(self):
    x = jnp.array([1.2, 3.0, 0.7])
    w = jnp.array([1.0, 2.0, 3.0])
    alpha = jnp.zeros_like(x)
    beta = jnp.ones_like(x) * 0.8
    c = 1.0
    params = (alpha, beta, w, c)
    p = projection.projection_box_section(x, params)
    self.assertAllClose(jnp.dot(w, p), c, atol=1e-5)
    self.assertTrue(jnp.all(alpha <= p))
    self.assertTrue(jnp.all(p <= beta))

    Q = jnp.eye(len(x))
    A = jnp.array([w])
    b = jnp.array([c])
    G = jnp.concatenate((-jnp.eye(len(x)), jnp.eye(len(x))))
    h = jnp.concatenate((alpha, beta))
    hyperparams = dict(params_obj=(Q, -x), params_eq=(A, b), params_ineq=(G, h))
    osqp = OSQP()
    p2 = osqp.run(**hyperparams).params.primal
    self.assertArraysAllClose(p, p2, atol=1e-3)

  def test_projection_box_section_infeasible(self):
    x = jnp.array([1.2, 3.0, 0.7])
    w = jnp.array([1.0, 2.0, 3.0])
    alpha = jnp.zeros_like(x)
    beta = jnp.ones_like(x)
    params = (alpha + 1.5, beta, w, 1.0)
    self.assertRaises(ValueError, projection.projection_box_section, x, params,
                      True)
    params = (alpha, beta - 1.5, w, 1.0)
    self.assertRaises(ValueError, projection.projection_box_section, x, params,
                      True)

  def test_projection_box_section_simplex(self):
    rng = onp.random.RandomState(0)
    x = jnp.array(rng.rand(5).astype(onp.float32))
    w = jnp.ones_like(x)
    alpha = jnp.zeros_like(x)
    beta = jnp.ones_like(x)
    c = 1.0
    params = (alpha, beta, w, c)

    p = projection.projection_simplex(x)
    p2 = projection.projection_box_section(x, params)
    self.assertArraysAllClose(p, p2, atol=1e-4, check_dtypes=False)

    J = jax.jacrev(projection.projection_simplex)(x)
    J2 = jax.jacrev(projection.projection_box_section)(x, params)
    self.assertArraysAllClose(J, J2, check_dtypes=False)

  def _sinkhorn(self, cost_matrix, marginals_a, marginals_b, maxiter=1000):
    eps = 1.0
    K = onp.exp(-cost_matrix / eps)
    f = onp.zeros_like(marginals_a)
    for it in range(maxiter):
      exp_f = onp.exp(f / eps)
      g = eps * onp.log(marginals_b) - eps * onp.log(K.T.dot(exp_f))
      exp_g = onp.exp(g / eps)
      f = eps * onp.log(marginals_a) - eps * onp.log(K.dot(exp_g))
    exp_f = onp.exp(f / eps)
    return exp_f[:, onp.newaxis] * K * exp_g

  def test_projection_transport(self):
    jax.config.update("jax_enable_x64", True)
    rng = onp.random.RandomState(0)
    sim_matrix = rng.rand(5, 6)
    marginals_a = rng.rand(5)
    marginals_a /= jnp.sum(marginals_a)
    marginals_b = rng.rand(6)
    marginals_b /= jnp.sum(marginals_b)
    make_solver = lambda fun: LBFGS(
        fun=fun, tol=1e-12, maxiter=10000, linesearch="zoom")

    for use_semi_dual in [False, True]:
      T1 = projection.projection_transport(
          sim_matrix, (marginals_a, marginals_b), make_solver,
          use_semi_dual=use_semi_dual)
      T2 = projection.kl_projection_transport(
          sim_matrix, (marginals_a, marginals_b), make_solver,
          use_semi_dual=use_semi_dual)

      for T in (T1, T2):
        self.assertEqual(T.shape, (5, 6))
        self.assertArraysAllClose(jnp.sum(T, axis=1), marginals_a, atol=1e-3)
        self.assertArraysAllClose(jnp.sum(T, axis=0), marginals_b, atol=1e-5)
        self.assertAllClose(jnp.sum(T), 1.0, atol=1e-12)

      # Compare with Sinkhorn.
      T3 = self._sinkhorn(cost_matrix=-sim_matrix, marginals_a=marginals_a,
                          marginals_b=marginals_b)
      self.assertArraysAllClose(T2, T3, atol=1e-4)

  def test_projection_birkhoff(self):
    rng = onp.random.RandomState(0)

    sim_matrix = rng.rand(5, 5)
    sim_matrix = 0.5 * (sim_matrix + sim_matrix.T)
    P1 = projection.projection_birkhoff(sim_matrix)
    P2 = projection.kl_projection_birkhoff(sim_matrix)

    # Check that P is a doubly-stochastic matrix.
    for P in (P1, P2):
      self.assertEqual(P.shape, (5, 5))
      self.assertArraysAllClose(jnp.sum(P, axis=1), jnp.ones(5), atol=1e-3)
      self.assertArraysAllClose(jnp.sum(P, axis=0), jnp.ones(5))

    # Check that input = output if input is already a doubly stochastic matrix.
    doubly_stochastic_matrix = jnp.array([[0.6, 0.4], [0.4, 0.6]])
    solution1 = projection.projection_birkhoff(doubly_stochastic_matrix)
    self.assertArraysAllClose(doubly_stochastic_matrix, solution1)

  def test_projection_sparse_simplex(self):
    def top_k(x, k):
      """Preserve the top-k entries of the vector x and put -inf values elsewhere.

      Args:
        x: (N,) array.
        k: integer.
      Returns:
        A array that preserves the top-k entries of and 0 elsewhere
      """

      _, top_k_indices = jax.lax.top_k(x, k)
      log_mask = jnp.log(jax.nn.one_hot(top_k_indices, len(x), dtype=x.dtype))
      return jnp.max(x[None, :] + log_mask, axis=0)

    dim_data = 100
    random_data = jax.random.normal(jax.random.PRNGKey(0), (dim_data,))

    # Check that the forward passes of projection_sparse_simplex_exact
    # and projection_simplex_sparse_approx_topk give correct results.
    for value in [1.0, 2.0]:
      ground_truth_composition = lambda x: projection.projection_simplex(  ## pylint: disable=g-long-lambda
          top_k(x, k=10), value=value)
      projection_sparse_simplex_exact = (
          lambda x: projection.projection_sparse_simplex(  ## pylint: disable=g-long-lambda
              x, max_nz=10, value=value, use_approx_max_nz=False))
      projection_sparse_simplex_approx = (
          lambda x: projection.projection_sparse_simplex(  ## pylint: disable=g-long-lambda
              x, max_nz=10, value=value, use_approx_max_nz=True))

      self.assertArraysAllClose(
          projection_sparse_simplex_exact(random_data),
          ground_truth_composition(random_data))
      self.assertArraysAllClose(
          projection_sparse_simplex_approx(random_data),
          ground_truth_composition(random_data))

      # Check that the backward passes of projection_sparse_simplex_exact
      # and projection_sparse_simplex_approx give correct results.
      self.assertArraysAllClose(
          jax.jacobian(projection_sparse_simplex_exact)(random_data),
          jax.jacobian(ground_truth_composition)(random_data))

      self.assertArraysAllClose(
          jax.jacobian(projection_sparse_simplex_approx)(random_data),
          jax.jacobian(ground_truth_composition)(random_data))
if __name__ == '__main__':
  absltest.main()



import math
import warnings
from contextlib import nullcontext as does_not_raise

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import scipy as sp
import scipy.stats as sts
import statsmodels.api as sm

import nemos as nmo
from nemos._observation_model_builder import (
    AVAILABLE_OBSERVATION_MODELS,
    instantiate_observation_model,
)
from nemos.glm.initialize_parameters import initialize_intercept_matching_mean_rate
from nemos.glm.inverse_link_function_utils import LINK_NAME_TO_FUNC


@pytest.fixture
def observation_model_rate_and_samples(observation_model_string, shape=None):
    """
    Fixture that returns rate and samples for each observation model.
    """
    if shape is None:
        shape = (10,)
    obs = instantiate_observation_model(observation_model_string)
    rate = jax.random.uniform(
        jax.random.PRNGKey(122), shape=shape, minval=0.1, maxval=10
    )
    if observation_model_string == "Poisson":
        y = jax.random.poisson(jax.random.PRNGKey(123), rate)
    elif observation_model_string == "Gamma":
        theta = 3
        y = jax.random.gamma(jax.random.PRNGKey(123), rate / theta) * theta
    elif observation_model_string == "Bernoulli":
        rate = rate / (1 + jnp.max(rate))
        y = jax.random.bernoulli(jax.random.PRNGKey(123), rate)
    elif observation_model_string == "NegativeBinomial":
        r = 1.0
        gamma_key, poisson_key = jax.random.split(jax.random.PRNGKey(123))
        gamma_sample = jax.random.gamma(gamma_key, r, shape=rate.shape) * (rate / r)
        y = jax.random.poisson(poisson_key, gamma_sample)
    else:
        raise ValueError(f"Unknown observation model {observation_model_string}.")
    return obs, y, rate


@pytest.fixture()
def poisson_observations():
    return nmo.observation_models.PoissonObservations


@pytest.fixture()
def gamma_observations():
    return nmo.observation_models.GammaObservations


@pytest.fixture()
def bernoulli_observations():
    return nmo.observation_models.BernoulliObservations


@pytest.fixture()
def negative_binomial_observations():
    return nmo.observation_models.NegativeBinomialObservations


@pytest.mark.parametrize(
    "obs_model_string, expectation",
    [
        ("Poisson", does_not_raise()),
        ("Gamma", does_not_raise()),
        ("Bernoulli", does_not_raise()),
        ("NegativeBinomial", does_not_raise()),
        (
            "invalid",
            pytest.raises(ValueError, match="Unknown observation model: invalid"),
        ),
    ],
)
@pytest.mark.parametrize("glm_class", [nmo.glm.GLM, nmo.glm.PopulationGLM])
def test_glm_instantiation_from_string_at_init(
    obs_model_string, glm_class, expectation
):
    with expectation:
        glm_class(observation_model=obs_model_string)


@pytest.mark.parametrize(
    "obs_model_string, expectation",
    [
        ("Poisson", does_not_raise()),
        ("Gamma", does_not_raise()),
        ("Bernoulli", does_not_raise()),
        ("nemos.observation_models.PoissonObservations", does_not_raise()),
        ("nemos.observation_models.GammaObservations", does_not_raise()),
        ("nemos.observation_models.BernoulliObservations", does_not_raise()),
        ("NegativeBinomial", does_not_raise()),
        ("nemos.observation_models.NegativeBinomial", does_not_raise()),
        (
            "invalid",
            pytest.raises(ValueError, match="Unknown observation model: invalid"),
        ),
    ],
)
@pytest.mark.parametrize("glm_class", [nmo.glm.GLM, nmo.glm.PopulationGLM])
def test_glm_setter_observation_model(obs_model_string, glm_class, expectation):
    """Test that the observation model can be set after the init providing a string."""
    if obs_model_string != "Poisson":
        obs = nmo.observation_models.PoissonObservations()
    else:
        obs = nmo.observation_models.GammaObservations()
    model = glm_class(observation_model=obs)
    with expectation:
        model.observation_model = obs_model_string
    if (
        obs_model_string == "Gamma"
        or obs_model_string == "nemos.observation_models.GammaObservations"
    ):
        assert isinstance(
            model.observation_model, nmo.observation_models.GammaObservations
        )
    elif (
        obs_model_string == "Poisson"
        or obs_model_string == "nemos.observation_models.PoissonObservations"
    ):
        assert isinstance(
            model.observation_model, nmo.observation_models.PoissonObservations
        )
    elif (
        obs_model_string == "Bernoulli"
        or obs_model_string == "nemos.observation_models.BernoulliObservations"
    ):
        assert isinstance(
            model.observation_model, nmo.observation_models.BernoulliObservations
        )
    elif obs_model_string == "NegativeBinomial":
        assert isinstance(
            model.observation_model, nmo.observation_models.NegativeBinomialObservations
        )


@pytest.mark.parametrize(
    "obs_model_string, expectation",
    [
        ("Poisson", does_not_raise()),
        ("Gamma", does_not_raise()),
        ("Bernoulli", does_not_raise()),
        ("NegativeBinomial", does_not_raise()),
        ("NB", pytest.raises(ValueError, match="Unknown observation model: NB")),
    ],
)
def test_instantiate_observation_model(obs_model_string, expectation):
    """Test instantiation of observation model with a link function."""
    with expectation:
        obs_model = instantiate_observation_model(
            obs_model_string,
        )


class TestPoissonObservations:

    def test_get_params(self, poisson_observations):
        """Test get_params() returns expected values."""
        observation_model = poisson_observations()

        assert observation_model.get_params() == {}

    def test_deviance_against_statsmodels(self, poissonGLM_model_instantiation):
        """
        Compare fitted parameters to statsmodels.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = poissonGLM_model_instantiation
        dev = sm.families.Poisson().deviance(y, firing_rate)
        dev_model = model.observation_model.deviance(y, firing_rate).sum()
        if not np.allclose(dev, dev_model):
            raise ValueError("Deviance doesn't match statsmodels!")

    def test_loglikelihood_against_scipy(self, poissonGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = poissonGLM_model_instantiation
        ll_model = model.observation_model.log_likelihood(y, firing_rate)
        ll_scipy = sts.poisson(firing_rate).logpmf(y).mean()
        if not np.allclose(ll_model, ll_scipy):
            raise ValueError("Log-likelihood doesn't match scipy!")

    def test_emission_probability(selfself, poissonGLM_model_instantiation):
        """
        Test the poisson emission probability.

        Check that the emission probability is set to jax.random.poisson.
        """
        _, _, model, _, _ = poissonGLM_model_instantiation
        key_array = jax.random.key(123)
        counts = model.observation_model.sample_generator(key_array, np.arange(1, 11))
        if not jnp.all(counts == jax.random.poisson(key_array, np.arange(1, 11))):
            raise ValueError(
                "The emission probability should output the results of a call to jax.random.poisson."
            )

    @pytest.mark.parametrize(
        "score_type, expectation",
        [
            ("pseudo-r2-McFadden", does_not_raise()),
            (
                "not-implemented",
                pytest.raises(
                    NotImplementedError, match="Score not-implemented not implemented"
                ),
            ),
        ],
    )
    def test_not_implemented_score(
        self, score_type, expectation, poissonGLM_model_instantiation
    ):
        _, y, model, _, firing_rate = poissonGLM_model_instantiation
        with expectation:
            model.observation_model.pseudo_r2(y, firing_rate, score_type)

    @pytest.mark.parametrize(
        "scale, expectation",
        [
            (1, does_not_raise()),
            (
                "invalid",
                pytest.raises(
                    ValueError, match="The `scale` parameter must be of numeric type"
                ),
            ),
        ],
    )
    def test_scale_setter(self, scale, expectation, poissonGLM_model_instantiation):
        _, _, model, _, firing_rate = poissonGLM_model_instantiation
        with expectation:
            model.observation_model.scale = scale

    def test_scale_getter(self, poissonGLM_model_instantiation):
        _, _, model, _, firing_rate = poissonGLM_model_instantiation
        assert model.observation_model.scale == 1

    @pytest.mark.parametrize(
        "test_value",
        [
            jnp.array([25.0]),
            jnp.array([7]),
            jnp.array([0.0]),
            jnp.array([1.0]),
            jnp.array([0.5]),
        ],
    )
    def test_custom_inverse_link_function(
        self,
        test_value,
        poissonGLM_model_instantiation,
    ):
        """
        Test that custom inverse link function can be inverted and works correctly.
        """
        # Initialize model
        _, _, model, _, _ = poissonGLM_model_instantiation
        model.observation_model.inverse_link_function = lambda x: jnp.power(x, 2)

        # Validate custom link function
        expected_output = jnp.sqrt(test_value)
        result = initialize_intercept_matching_mean_rate(
            model.observation_model.inverse_link_function, test_value
        )

        assert np.allclose(
            result, expected_output
        ), f"Inverse link function result mismatch: expected {expected_output}, got {result}"

    def test_pseudo_r2_vs_statsmodels(self, poissonGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        X, y, model, _, firing_rate = poissonGLM_model_instantiation

        # statsmodels mcfadden
        mdl = sm.GLM(y, sm.add_constant(X), family=sm.families.Poisson()).fit()
        pr2_sms = mdl.pseudo_rsquared("mcf")

        # set params
        pr2_model = model.observation_model.pseudo_r2(
            y, mdl.mu, score_type="pseudo-r2-McFadden"
        )

        if not np.allclose(pr2_model, pr2_sms):
            raise ValueError("Log-likelihood doesn't match statsmodels!")

    def test_repr_out(self):
        obs = nmo.observation_models.PoissonObservations()
        assert repr(obs) == f"PoissonObservations()"


class TestGammaObservations:

    def test_get_params(self, gamma_observations):
        """Test get_params() returns expected values."""
        observation_model = gamma_observations()

        assert observation_model.get_params() == {}

    def test_deviance_against_statsmodels(self, gammaGLM_model_instantiation):
        """
        Compare fitted parameters to statsmodels.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = gammaGLM_model_instantiation
        dev = sm.families.Gamma().deviance(y, firing_rate)
        dev_model = model.observation_model.deviance(y, firing_rate).sum()
        if not np.allclose(dev, dev_model):
            raise ValueError("Deviance doesn't match statsmodels!")

    def test_loglikelihood_against_statsmodels(self, gammaGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = gammaGLM_model_instantiation
        ll_model = model.observation_model.log_likelihood(y, firing_rate)
        ll_sms = sm.families.Gamma().loglike(y, firing_rate) / y.shape[0]
        if not np.allclose(ll_model, ll_sms):
            raise ValueError("Log-likelihood doesn't match statsmodels!")

    def test_emission_probability(self, gammaGLM_model_instantiation):
        """
        Test the gamma emission probability.

        Check that the emission probability is set to jax.random.gamma.
        """
        _, _, model, _, _ = gammaGLM_model_instantiation
        key_array = jax.random.key(123)
        counts = model.observation_model.sample_generator(key_array, np.arange(1, 11))
        if not jnp.all(counts == jax.random.gamma(key_array, np.arange(1, 11))):
            raise ValueError(
                "The emission probability should output the results of a call to jax.random.gamma."
            )

    def test_pseudo_r2_vs_statsmodels(self, gammaGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        jax.config.update("jax_enable_x64", True)
        X, y, model, _, firing_rate = gammaGLM_model_instantiation

        # statsmodels mcfadden
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="The InversePower link function does"
            )
            mdl = sm.GLM(y, sm.add_constant(X), family=sm.families.Gamma()).fit()
        pr2_sms = mdl.pseudo_rsquared("mcf")

        # set params
        pr2_model = model.observation_model.pseudo_r2(
            y, mdl.mu, score_type="pseudo-r2-McFadden", scale=mdl.scale
        )

        if not np.allclose(pr2_model, pr2_sms):
            raise ValueError("Log-likelihood doesn't match statsmodels!")

    def test_repr_out(self):
        obs = nmo.observation_models.GammaObservations()
        assert repr(obs) == f"GammaObservations()"


class TestBernoulliObservations:

    def test_get_params(self, bernoulli_observations):
        """Test get_params() returns expected values."""
        observation_model = bernoulli_observations()

        assert observation_model.get_params() == {}

    def test_deviance_against_statsmodels(self, bernoulliGLM_model_instantiation):
        """
        Compare fitted parameters to statsmodels.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = bernoulliGLM_model_instantiation
        dev = sm.families.Binomial().deviance(y, firing_rate)
        dev_model = model.observation_model.deviance(y, firing_rate).sum()
        if not np.allclose(dev, dev_model):
            raise ValueError("Deviance doesn't match statsmodels!")

    def test_loglikelihood_against_scipy(self, bernoulliGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = bernoulliGLM_model_instantiation
        ll_model = model.observation_model.log_likelihood(y, firing_rate)
        ll_scipy = sts.bernoulli(firing_rate).logpmf(y).mean()
        if not np.allclose(ll_model, ll_scipy):
            raise ValueError("Log-likelihood doesn't match scipy!")

    def test_emission_probability(self, bernoulliGLM_model_instantiation):
        """
        Test the poisson emission probability.

        Check that the emission probability is set to jax.random.poisson.
        """
        _, _, model, _, _ = bernoulliGLM_model_instantiation
        key_array = jax.random.key(123)
        p = np.random.rand(10)
        counts = model.observation_model.sample_generator(key_array, p)
        if not jnp.all(counts == jax.random.bernoulli(key_array, p)):
            raise ValueError(
                "The emission probability should output the results of a call to jax.random.poisson."
            )

    def test_pseudo_r2_vs_statsmodels(self, bernoulliGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        X, y, model, _, firing_rate = bernoulliGLM_model_instantiation

        # statsmodels mcfadden
        mdl = sm.GLM(y, sm.add_constant(X), family=sm.families.Binomial()).fit()
        pr2_sms = mdl.pseudo_rsquared("mcf")

        # set params
        pr2_model = model.observation_model.pseudo_r2(
            y, mdl.mu, score_type="pseudo-r2-McFadden"
        )

        if not np.allclose(pr2_model, pr2_sms):
            raise ValueError("Log-likelihood doesn't match statsmodels!")

    def test_repr_out(self):
        obs = nmo.observation_models.BernoulliObservations()
        assert repr(obs) == f"BernoulliObservations()"


class TestNegativeBinomialObservations:

    def test_get_params(self, negative_binomial_observations):
        observation_model = negative_binomial_observations()
        assert observation_model.get_params() == {
            "scale": 1.0,
        }

    def test_deviance_against_statsmodels(
        self, negativeBinomialGLM_model_instantiation
    ):
        jax.config.update("jax_enable_x64", True)
        _, y, model, _, firing_rate = negativeBinomialGLM_model_instantiation
        dev = sm.families.NegativeBinomial(
            alpha=model.observation_model.scale
        ).deviance(y, firing_rate)
        dev_model = model.observation_model.deviance(y, firing_rate).sum()
        if not np.allclose(dev, dev_model):
            raise ValueError("Deviance doesn't match statsmodels!")

    def test_loglikelihood_against_scipy(self, negativeBinomialGLM_model_instantiation):
        _, y, model, _, firing_rate = negativeBinomialGLM_model_instantiation
        r = 1.0 / model.observation_model.scale
        p = r / (r + firing_rate)
        ll_model = model.observation_model.log_likelihood(y, firing_rate)
        ll_scipy = sts.nbinom.logpmf(y, r, p).mean()
        if not np.allclose(ll_model, ll_scipy, atol=1e-5):
            raise ValueError("Log-likelihood doesn't match scipy!")

    def test_emission_probability(self, negativeBinomialGLM_model_instantiation):
        _, _, model, _, _ = negativeBinomialGLM_model_instantiation
        key_array = jax.random.key(123)
        gkey, pkey = jax.random.split(key_array)
        p = np.random.rand(10)
        counts = model.observation_model.sample_generator(
            key_array, p, scale=model.observation_model.scale
        )
        r = 1.0 / model.observation_model.scale
        gamma_sample = jax.random.gamma(gkey, r, shape=p.shape) * (p / r)
        expected_counts = jax.random.poisson(pkey, gamma_sample)
        if not jnp.allclose(counts, expected_counts):
            raise ValueError(
                "The emission probability doesn't match expected NB sampling."
            )

    def test_pseudo_r2_vs_statsmodels(self, negativeBinomialGLM_model_instantiation):
        X, y, model, _, firing_rate = negativeBinomialGLM_model_instantiation
        mdl = sm.GLM(
            y,
            sm.add_constant(X),
            family=sm.families.NegativeBinomial(alpha=model.observation_model.scale),
        ).fit()
        pr2_sms = mdl.pseudo_rsquared("mcf")
        pr2_model = model.observation_model.pseudo_r2(
            y,
            mdl.mu,
            scale=model.observation_model.scale,
            score_type="pseudo-r2-McFadden",
        )
        if not np.allclose(pr2_model, pr2_sms, atol=1e-5):
            raise ValueError("Pseudo-r2 doesn't match statsmodels!")

    def test_repr_out(self):
        obs = nmo.observation_models.NegativeBinomialObservations()
        assert repr(obs) == f"NegativeBinomialObservations(scale=1.0)"


@pytest.mark.parametrize("observation_model_string", AVAILABLE_OBSERVATION_MODELS)
class TestCommonObservationModels:

    @pytest.mark.parametrize("shape", [(10,), (10, 5), (10, 5, 2)])
    def test_likelihood_matching(
        self, shape, observation_model_string, observation_model_rate_and_samples
    ):
        jax.config.update("jax_enable_x64", True)
        obs, y, rate = observation_model_rate_and_samples
        like1 = jnp.exp(
            obs.log_likelihood(y, rate, aggregate_sample_scores=lambda x: x)
        )
        like2 = obs.likelihood(y, rate, aggregate_sample_scores=lambda x: x)
        assert jnp.allclose(like1, like2)

        like1 = jnp.exp(obs.log_likelihood(y, rate))
        like2 = obs.likelihood(y, rate)
        assert jnp.allclose(like1, like2)

    @pytest.mark.parametrize("shape", [(10,), (10, 5), (10, 5, 2)])
    def test_aggregation_score_mcfadden(
        self, shape, observation_model_string, observation_model_rate_and_samples
    ):
        obs, y, rate = observation_model_rate_and_samples
        sm = obs._pseudo_r2_mcfadden(y, rate, aggregate_sample_scores=jnp.sum)
        mn = obs._pseudo_r2_mcfadden(y, rate, aggregate_sample_scores=jnp.mean)
        assert np.allclose(sm, mn)

    @pytest.mark.parametrize("shape", [(10,), (10, 5), (10, 5, 2)])
    def test_aggregation_score_choen(
        self, shape, observation_model_string, observation_model_rate_and_samples
    ):
        obs, y, rate = observation_model_rate_and_samples
        sm = obs._pseudo_r2_cohen(y, rate, aggregate_sample_scores=jnp.sum)
        mn = obs._pseudo_r2_cohen(y, rate, aggregate_sample_scores=jnp.mean)
        assert np.allclose(sm, mn)

    @pytest.mark.parametrize("score_type", ["pseudo-r2-McFadden", "pseudo-r2-Cohen"])
    @pytest.mark.parametrize("shape", [(10,), (10, 5), (10, 5, 2)])
    def test_aggregation_score_pr2(
        self,
        score_type,
        shape,
        observation_model_string,
        observation_model_rate_and_samples,
    ):
        obs, y, rate = observation_model_rate_and_samples
        sm = obs.pseudo_r2(
            y, rate, score_type=score_type, aggregate_sample_scores=jnp.sum
        )
        mn = obs.pseudo_r2(
            y, rate, score_type=score_type, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn)

    @pytest.mark.parametrize("shape", [(10,), (10, 5), (10, 5, 2)])
    def test_aggregation_score_ll(
        self, shape, observation_model_string, observation_model_rate_and_samples
    ):
        obs, y, rate = observation_model_rate_and_samples
        sm = obs.log_likelihood(y, rate, aggregate_sample_scores=jnp.sum)
        mn = obs.log_likelihood(y, rate, aggregate_sample_scores=jnp.mean)
        assert np.allclose(sm, mn * math.prod(y.shape))

    @pytest.mark.parametrize("shape", [(10,), (10, 5), (10, 5, 2)])
    def test_aggregation_score_neg_ll(
        self, shape, observation_model_string, observation_model_rate_and_samples
    ):
        obs, y, rate = observation_model_rate_and_samples
        sm = obs._negative_log_likelihood(y, rate, jnp.sum)
        mn = obs._negative_log_likelihood(y, rate, jnp.mean)
        assert np.allclose(sm, mn * math.prod(y.shape))

    def test_scale_getter(
        self, observation_model_string, observation_model_rate_and_samples
    ):
        obs, y, rate = observation_model_rate_and_samples
        assert obs.scale == 1

    @pytest.mark.parametrize(
        "scale, expectation",
        [
            (1, does_not_raise()),
            (
                "invalid",
                pytest.raises(
                    ValueError, match="The `scale` parameter must be of numeric type"
                ),
            ),
        ],
    )
    def test_scale_setter(
        self,
        scale,
        expectation,
        observation_model_string,
        observation_model_rate_and_samples,
    ):
        obs, y, rate = observation_model_rate_and_samples
        with expectation:
            obs.scale = scale

    @pytest.mark.parametrize(
        "score_type, expectation",
        [
            ("pseudo-r2-McFadden", does_not_raise()),
            (
                "not-implemented",
                pytest.raises(
                    NotImplementedError, match="Score not-implemented not implemented"
                ),
            ),
        ],
    )
    def test_not_implemented_score(
        self,
        score_type,
        expectation,
        observation_model_string,
        observation_model_rate_and_samples,
    ):
        obs, y, rate = observation_model_rate_and_samples
        with expectation:
            obs.pseudo_r2(y, rate, score_type)

    @pytest.mark.parametrize("score_type", ["pseudo-r2-Cohen", "pseudo-r2-McFadden"])
    def test_pseudo_r2_mean(
        self, score_type, observation_model_string, observation_model_rate_and_samples
    ):
        """
        Check that the pseudo-r2 of the null model is 0.
        """
        obs, y, rate = observation_model_rate_and_samples
        pseudo_r2 = obs.pseudo_r2(y, y.mean(), score_type=score_type)
        if not np.allclose(pseudo_r2, 0, atol=10**-7, rtol=0.0):
            raise ValueError(
                f"pseudo-r2 of {pseudo_r2} for the null model. Should be equal to 0!"
            )

    @pytest.mark.parametrize("score_type", ["pseudo-r2-Cohen", "pseudo-r2-McFadden"])
    def test_pseudo_r2_range(
        self, score_type, observation_model_string, observation_model_rate_and_samples
    ):
        """
        Compute the pseudo-r2 and check that is < 1.
        """
        obs, y, rate = observation_model_rate_and_samples
        pseudo_r2 = obs.pseudo_r2(y, rate, score_type=score_type)
        if (pseudo_r2 > 1) or (pseudo_r2 < 0):
            raise ValueError(f"pseudo-r2 of {pseudo_r2} outside the [0,1] range!")


class ConfigurableObservationModel:
    """
    A configurable observation model for testing.

    Parameters
    ----------
    exclude_methods : list of str, optional
        Methods to exclude from the model (simulates missing attributes)
    override_returns : dict, optional
        Dictionary mapping method names to return values.
        If a method is in this dict, it will return the specified value instead of the default.
    non_callable_methods : list of str, optional
        Methods that should be non-callable (assigned as strings or other non-callable types)
    """

    def __init__(
        self, exclude_methods=None, override_returns=None, non_callable_methods=None
    ):
        self.exclude_methods = exclude_methods or []
        self.override_returns = override_returns or {}
        self.non_callable_methods = non_callable_methods or []

        # Set non-callable attributes
        for method_name in self.non_callable_methods:
            setattr(self, method_name, "not_a_function")

    def __getattribute__(self, name):
        # Use object's __getattribute__ to avoid infinite recursion
        exclude_methods = object.__getattribute__(self, "exclude_methods")

        # If method is excluded, raise AttributeError
        if name in exclude_methods:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )

        return object.__getattribute__(self, name)

    @property
    def default_inverse_link_function(self):
        if "default_inverse_link_function" in self.override_returns:
            return lambda x: self.override_returns["default_inverse_link_function"]
        return jax.scipy.special.expit

    def _negative_log_likelihood(self, y, rate, aggregate_sample_scores=jnp.mean):
        if "_negative_log_likelihood" in self.override_returns:
            return self.override_returns["_negative_log_likelihood"]
        return -aggregate_sample_scores(
            y * jax.scipy.special.logit(rate)
            + (1 - y) * jax.scipy.special.logit(1 - rate)
        )

    def pseudo_r2(self, y, rate, aggregate_sample_scores=jnp.mean):
        if "pseudo_r2" in self.override_returns:
            return self.override_returns["pseudo_r2"]
        return 1 - (
            self._negative_log_likelihood(y, rate, aggregate_sample_scores)
            / jnp.sum((y - y.mean()) ** 2)
        )

    def sample_generator(self, key, rate, scale=1.0):
        if "sample_generator" in self.override_returns:
            return self.override_returns["sample_generator"]
        return jax.random.bernoulli(key, rate)

    def estimate_scale(self, y, predicted_rate, dof_resid):
        if "estimate_scale" in self.override_returns:
            return self.override_returns["estimate_scale"]
        return jnp.array(1.0)

    def log_likelihood(self, y, rate, aggregate_sample_scores=jnp.mean):
        if "log_likelihood" in self.override_returns:
            return self.override_returns["log_likelihood"]
        return -self._negative_log_likelihood(y, rate, aggregate_sample_scores)

    def likelihood(self, y, rate, aggregate_sample_scores=jnp.mean):
        if "likelihood" in self.override_returns:
            return self.override_returns["likelihood"]
        return jnp.exp(self.log_likelihood(y, rate, aggregate_sample_scores))

    def deviance(self, y, rate, scale=1.0):
        if "deviance" in self.override_returns:
            return self.override_returns["deviance"]
        identity = lambda x: x
        return 2 * (
            self.log_likelihood(y, rate, identity)
            - self.log_likelihood(y.mean() * np.ones_like(y), rate, identity)
        )


@pytest.fixture
def valid_observation_model():
    """A valid observation model that passes all checks."""
    return ConfigurableObservationModel()


def test_valid_observation_model(valid_observation_model):
    """Test that a valid observation model passes all checks."""
    nmo.observation_models.check_observation_model(valid_observation_model)


@pytest.mark.parametrize("observation", AVAILABLE_OBSERVATION_MODELS)
def test_nemos_model_pass_check(observation):
    """Test that a valid observation model passes all checks."""
    obs = instantiate_observation_model(observation)
    nmo.observation_models.check_observation_model(obs, force_checks=True)


@pytest.mark.parametrize(
    "missing_method",
    [
        "_negative_log_likelihood",
        "pseudo_r2",
        "log_likelihood",
        "likelihood",
        "deviance",
        "sample_generator",
        "estimate_scale",
        "default_inverse_link_function",
    ],
)
def test_missing_method(missing_method):
    """Test that missing required methods raise AttributeError."""
    model = ConfigurableObservationModel(exclude_methods=[missing_method])

    with pytest.raises(AttributeError, match="does not have the required"):
        nmo.observation_models.check_observation_model(model)


@pytest.mark.parametrize(
    "non_callable_method",
    [
        "_negative_log_likelihood",
        "pseudo_r2",
        "log_likelihood",
        "likelihood",
        "deviance",
        "sample_generator",
        "estimate_scale",
    ],
)
def test_non_callable_method(non_callable_method):
    """Test that non-callable methods raise TypeError."""
    model = ConfigurableObservationModel(non_callable_methods=[non_callable_method])

    with pytest.raises(TypeError, match="must be a Callable"):
        nmo.observation_models.check_observation_model(model)


@pytest.mark.parametrize(
    "method_name,return_value",
    [
        ("_negative_log_likelihood", "not an array"),
        ("pseudo_r2", [1, 2, 3]),
        ("log_likelihood", {"key": "value"}),
        ("likelihood", 42),
        ("deviance", "string"),
        ("sample_generator", [1.0, 2.0, 3.0]),
        ("estimate_scale", None),
    ],
)
def test_method_not_returning_array(method_name, return_value):
    """Test that methods not returning ndarray raise TypeError."""
    model = ConfigurableObservationModel(override_returns={method_name: return_value})

    with pytest.raises(TypeError, match="must return a"):
        nmo.observation_models.check_observation_model(model)


@pytest.mark.parametrize(
    "method_name,return_value",
    [
        ("_negative_log_likelihood", jnp.array([1.0, 2.0, 3.0])),
        ("pseudo_r2", jnp.array([0.5, 0.6, 0.7])),
        ("log_likelihood", jnp.array([[1.0], [2.0]])),
        ("likelihood", jnp.array([1.0, 1.0])),
    ],
)
def test_method_not_returning_scalar(method_name, return_value):
    """Test that methods marked as scalar functions return scalars, not vectors."""
    model = ConfigurableObservationModel(override_returns={method_name: return_value})

    with pytest.raises(TypeError, match="should return a scalar"):
        nmo.observation_models.check_observation_model(model)


@pytest.mark.parametrize(
    "return_value,description",
    [
        (jnp.array([1.0, 2.0]), "too few elements"),
        (jnp.array([1.0, 2.0, 3.0, 4.0]), "too many elements"),
        (jnp.array([[1.0, 2.0, 3.0]]), "wrong dimensionality"),
    ],
)
def test_sample_generator_wrong_shape(return_value, description):
    """Test that sample_generator preserves input shape."""
    model = ConfigurableObservationModel(
        override_returns={"sample_generator": return_value}
    )

    with pytest.raises(ValueError, match="preserve the input array"):
        nmo.observation_models.check_observation_model(model)
